from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphql import GraphQLError, Undefined, ValidationRule, ast_from_value
from graphql.language import ast

from undine import InterfaceType, MutationType, QueryType, UnionType
from undine.relay import Connection
from undine.utils.graphql.undine_extensions import (
    get_undine_calculation_argument,
    get_undine_directive,
    get_undine_directive_argument,
    get_undine_entrypoint,
    get_undine_field,
    get_undine_filter,
    get_undine_filterset,
    get_undine_input,
    get_undine_interface_field,
    get_undine_mutation_type,
    get_undine_order,
    get_undine_orderset,
    get_undine_query_type,
)
from undine.utils.graphql.utils import get_underlying_type
from undine.utils.reflection import is_subclass

if TYPE_CHECKING:
    from collections.abc import Generator

    from graphql import (
        FieldNode,
        GraphQLCompositeType,
        GraphQLDirective,
        GraphQLEnumType,
        GraphQLInputObjectType,
        GraphQLInputType,
        VisitorAction,
    )

    from undine import Entrypoint, Field, InterfaceField
    from undine.execution import UndineValidationContext


__all__ = [
    "VisibilityRule",
]


class VisibilityRule(ValidationRule):  # noqa: PLR0904
    """Validates that fields that are not visible to the user are not queried."""

    context: UndineValidationContext

    # Entry hooks

    def enter_field(self, node: ast.FieldNode, *args: Any) -> VisitorAction:
        parent_type = self.context.get_parent_type()
        if not parent_type:
            return None

        graphql_field = self.context.get_field_def()
        if not graphql_field:
            return None

        undine_entrypoint = get_undine_entrypoint(graphql_field)
        if undine_entrypoint is not None:
            return self.handle_entrypoint(undine_entrypoint, parent_type, node)

        undine_field = get_undine_field(graphql_field)
        if undine_field is not None:
            return self.handle_field(undine_field, parent_type, node)

        undine_interface_field = get_undine_interface_field(graphql_field)
        if undine_interface_field is not None:
            return self.handle_interface_field(undine_interface_field, parent_type, node)

        return None

    def enter_argument(self, node: ast.ArgumentNode, *args: Any) -> VisitorAction:  # noqa: PLR0911,PLR0912,C901
        # Get last ancestor, which is the field node containing the argument.
        field_node: FieldNode = args[-1][-1]

        graphql_argument = self.context.get_argument()
        if graphql_argument is None:
            return None

        parent_type = self.context.get_parent_type()
        if not parent_type:
            return None

        graphql_input_type = self.context.get_input_type()
        if graphql_input_type is None:
            return None

        node_value = node.value
        if isinstance(node_value, ast.VariableNode):
            node_value = self.context.variable_as_ast(node_value.name.value, graphql_input_type)
            if node_value is None:
                return None

        while hasattr(graphql_input_type, "of_type"):
            graphql_input_type = graphql_input_type.of_type

        undine_filterset = get_undine_filterset(graphql_input_type)
        if undine_filterset is not None:
            if not undine_filterset.__is_visible__(self.context.request):
                self.report_field_argument_error(parent_type, field_node, node)
                return self.BREAK

            filterset_node_value: ast.ObjectValueNode = node_value  # type: ignore[assignment]
            return self.handle_filters(graphql_input_type, filterset_node_value)

        undine_orderset = get_undine_orderset(graphql_input_type)
        if undine_orderset is not None:
            if not undine_orderset.__is_visible__(self.context.request):
                self.report_field_argument_error(parent_type, field_node, node)
                return self.BREAK

            orderset_node_value: ast.EnumValueNode | ast.ListValueNode = node_value  # type: ignore[assignment]
            return self.handle_orders(graphql_input_type, orderset_node_value)

        undine_mutation_type = get_undine_mutation_type(graphql_input_type)
        if undine_mutation_type is not None:
            if not undine_mutation_type.__is_visible__(self.context.request):
                self.report_field_argument_error(parent_type, field_node, node)
                return self.BREAK

            mutation_node_value: ast.ObjectValueNode | ast.ListValueNode = node_value  # type: ignore[assignment]
            return self.handle_inputs(graphql_input_type, mutation_node_value)

        undine_calculation_arg = get_undine_calculation_argument(graphql_argument)
        if undine_calculation_arg is not None:
            if undine_calculation_arg.visible_func is None:
                return None

            if not undine_calculation_arg.visible_func(undine_calculation_arg, self.context.request):
                self.report_field_argument_error(parent_type, field_node, node)
                return self.BREAK

            return None

        graphql_directive = self.context.get_directive()
        if graphql_directive is not None:
            return self.handle_directive_arguments(graphql_directive, node)

        return None

    def enter_named_type(self, node: ast.NamedTypeNode, *args: Any) -> VisitorAction:
        graphql_type = self.context.get_type()
        if graphql_type is None:
            # Handled by `graphql.validation.rules.known_type_names.KnownTypeNamesRule`
            return None

        # Check that fragment definitions and inline fragments can be used on this type.
        undine_query_type = get_undine_query_type(graphql_type)
        if undine_query_type is not None:
            if not undine_query_type.__is_visible__(self.context.request):
                self.report_type_error(graphql_type, node)
                return self.BREAK

            return None

        return None

    def enter_directive(self, node: ast.DirectiveNode, *args: Any) -> VisitorAction:
        graphql_directive = self.context.get_directive()
        if graphql_directive is None:
            return None

        undine_directive = get_undine_directive(graphql_directive)
        if undine_directive is None:
            return None

        if not undine_directive.__is_visible__(self.context.request):
            self.report_directive_error(graphql_directive, node)
            return self.BREAK

        return None

    # handle undine types

    def handle_entrypoint(
        self,
        undine_entrypoint: Entrypoint,
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if undine_entrypoint.visible_func is not None:
            if not undine_entrypoint.visible_func(undine_entrypoint, self.context.request):
                self.report_field_error(parent_type, field_node)
                return self.BREAK

            return None

        ref = undine_entrypoint.ref

        if isinstance(ref, Connection):
            if ref.query_type is not None:
                ref = ref.query_type
            elif ref.union_type is not None:
                ref = ref.union_type
            elif ref.interface_type is not None:
                ref = ref.interface_type

        if is_subclass(ref, QueryType):
            return self.handle_query_type(ref, parent_type, field_node)

        if is_subclass(ref, MutationType):
            return self.handle_mutation_type(ref, parent_type, field_node)

        if is_subclass(ref, InterfaceType):
            return self.handle_interface_type(ref, parent_type, field_node)

        if is_subclass(ref, UnionType):
            return self.handle_union_type(ref, parent_type, field_node)

        return None

    def handle_field(
        self,
        undine_field: Field,
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if undine_field.visible_func is not None:
            if not undine_field.visible_func(undine_field, self.context.request):
                self.report_field_error(parent_type, field_node)
                return self.BREAK

            return None

        ref = undine_field.ref

        if isinstance(ref, Connection) and ref.query_type is not None:
            ref = ref.query_type

        if is_subclass(ref, QueryType):
            return self.handle_query_type(ref, parent_type, field_node)

        if is_subclass(ref, MutationType):
            return self.handle_mutation_type(ref, parent_type, field_node)

        return None

    def handle_interface_field(
        self,
        undine_interface_field: InterfaceField,
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if undine_interface_field.visible_func is None:
            return None

        if not undine_interface_field.visible_func(undine_interface_field, self.context.request):
            self.report_field_error(parent_type, field_node)
            return self.BREAK

        return None

    def handle_query_type(
        self,
        ref: type[QueryType],
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if not ref.__is_visible__(self.context.request):
            self.report_field_error(parent_type, field_node)
            return self.BREAK

        return None

    def handle_mutation_type(
        self,
        ref: type[MutationType],
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if not ref.__is_visible__(self.context.request):
            self.report_field_error(parent_type, field_node)
            return self.BREAK

        output_type = ref.__output_type__()
        query_type = get_undine_query_type(output_type)
        if query_type is not None and not query_type.__is_visible__(self.context.request):
            self.report_field_error(parent_type, field_node)
            return self.BREAK

        return None

    def handle_interface_type(
        self,
        ref: type[InterfaceType],
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if not ref.__is_visible__(self.context.request):
            self.report_field_error(parent_type, field_node)
            return self.BREAK

        return None

    def handle_union_type(
        self,
        ref: type[UnionType],
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> VisitorAction:
        if not ref.__is_visible__(self.context.request):
            self.report_field_error(parent_type, field_node)
            return self.BREAK

        return None

    def handle_filters(
        self,
        input_type: GraphQLInputObjectType,
        node: ast.ObjectValueNode,
    ) -> VisitorAction:
        action: VisitorAction = None
        for field_node in self.flatten_filters(node.fields[0], input_type):
            if field_node is None:
                continue

            filter_name = field_node.name.value
            input_field = input_type.fields.get(filter_name)
            if input_field is None:
                continue

            undine_filter = get_undine_filter(input_field)
            if undine_filter is None:
                continue

            if undine_filter.visible_func is None:
                continue

            if not undine_filter.visible_func(undine_filter, self.context.request):
                self.report_input_field_error(input_type, field_node)
                action = self.BREAK

        return action

    def handle_orders(
        self,
        enum_type: GraphQLEnumType,
        node: ast.ListValueNode | ast.EnumValueNode,
    ) -> VisitorAction:
        action: VisitorAction = None

        if isinstance(node, ast.EnumValueNode):
            node = ast.ListValueNode(values=[node])

        value_node: ast.EnumValueNode | ast.VariableNode
        for value_node in node.values:
            if isinstance(value_node, ast.VariableNode):
                value = self.context.variables.get(value_node.name.value, Undefined)
                if value is Undefined:
                    continue

                value_node = ast.EnumValueNode(value=value)  # noqa: PLW2901

            enum_name = value_node.value
            enum_value = enum_type.values.get(enum_name)
            if enum_value is None:
                continue

            undine_order = get_undine_order(enum_value)
            if undine_order is None:
                continue

            if undine_order.visible_func is None:
                continue

            if not undine_order.visible_func(undine_order, self.context.request):
                self.report_enum_error(enum_type, value_node)
                action = self.BREAK

        return action

    def handle_inputs(
        self,
        input_type: GraphQLInputObjectType,
        node: ast.ObjectValueNode | ast.ListValueNode,
    ) -> VisitorAction:
        action: VisitorAction = None

        if isinstance(node, ast.ObjectValueNode):
            node = ast.ListValueNode(values=[node])

        item: ast.ObjectValueNode | ast.VariableNode
        for item in node.values:
            if isinstance(item, ast.VariableNode):
                item = self.context.variable_as_ast(item.name.value, input_type)  # type: ignore[assignment]  # noqa: PLW2901
                if item is None:
                    continue

            for field_node in item.fields:
                input_name = field_node.name.value
                input_field = input_type.fields.get(input_name)
                if input_field is None:
                    continue

                undine_input = get_undine_input(input_field)
                if undine_input is None:
                    continue

                if undine_input.visible_func is None:
                    undine_input_type: GraphQLInputObjectType = get_underlying_type(input_field.type)  # type: ignore[assignment]
                    undine_mutation_type = get_undine_mutation_type(undine_input_type)
                    if undine_mutation_type is None:
                        continue

                    if not undine_mutation_type.__is_visible__(self.context.request):
                        self.report_input_field_error(input_type, field_node)
                        action = self.BREAK

                    continue

                if not undine_input.visible_func(undine_input, self.context.request):
                    self.report_input_field_error(input_type, field_node)
                    action = self.BREAK

        return action

    def handle_directive_arguments(
        self,
        directive_type: GraphQLDirective,
        node: ast.ArgumentNode,
    ) -> VisitorAction:
        arg = directive_type.args.get(node.name.value)
        if arg is None:
            return None

        undine_directive_arg = get_undine_directive_argument(arg)
        if undine_directive_arg is None:
            return None

        if undine_directive_arg.visible_func is None:
            return None

        if not undine_directive_arg.visible_func(undine_directive_arg, self.context.request):
            self.report_directive_argument_error(directive_type, node)
            return self.BREAK

        return None

    # Report errors

    def report_type_error(
        self,
        parent_type: GraphQLCompositeType,
        node: ast.NamedTypeNode,
    ) -> None:
        # This type is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include types that are not visible.
        msg = f"Unknown type '{node.name.value}'."
        self.report_error(GraphQLError(msg, node))

    def report_field_error(
        self,
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
    ) -> None:
        # This field is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include fields that are not visible.
        msg = f"Cannot query field '{field_node.name.value}' on type '{parent_type}'."
        self.report_error(GraphQLError(msg, nodes=field_node))

    def report_field_argument_error(
        self,
        parent_type: GraphQLCompositeType,
        field_node: ast.FieldNode,
        arg_node: ast.ArgumentNode,
    ) -> None:
        # This argument is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include argument that are not visible.
        msg = f"Unknown argument '{arg_node.name.value}' on field '{parent_type}.{field_node.name.value}'."
        self.report_error(GraphQLError(msg, nodes=arg_node))

    def report_directive_argument_error(
        self,
        parent_type: GraphQLDirective,
        arg_node: ast.ArgumentNode,
    ) -> None:
        # This argument is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include argument that are not visible.
        msg = f"Unknown argument '{arg_node.name.value}' on directive '{parent_type}'."
        self.report_error(GraphQLError(msg, nodes=arg_node))

    def report_input_field_error(
        self,
        parent_type: GraphQLInputObjectType,
        object_field_node: ast.ObjectFieldNode,
    ) -> None:
        # This argument is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include arguments that are not visible.
        msg = f"Field '{object_field_node.name.value}' is not defined by type '{parent_type.name}'."
        self.report_error(GraphQLError(msg, nodes=object_field_node))

    def report_enum_error(
        self,
        parent_type: GraphQLEnumType,
        enum_value_node: ast.EnumValueNode,
    ) -> None:
        # This enum value is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include values that are not visible.
        msg = f"Value '{enum_value_node.value}' does not exist in '{parent_type.name}' enum."
        self.report_error(GraphQLError(msg, nodes=enum_value_node))

    def report_directive_error(
        self,
        parent_type: GraphQLDirective,
        directive_node: ast.DirectiveNode,
    ) -> None:
        # This directive is invisible so treat is as if it doesn't exist.
        # Do not include suggestions, since they might include directives that are not visible.
        msg = f"Unknown directive '@{directive_node.name.value}'."
        self.report_error(GraphQLError(msg, nodes=directive_node))

    # Helpers

    def flatten_filters(
        self,
        node: ast.ObjectFieldNode,
        input_type: GraphQLInputType,
    ) -> Generator[ast.ObjectFieldNode | None, None, None]:
        node_value = node.value

        if node.name.value in {"AND", "OR", "XOR", "NOT"} and isinstance(node_value, ast.ObjectValueNode):
            for sub_node in node_value.fields:
                field_type = input_type.fields[sub_node.name.value].type
                yield from self.flatten_filters(sub_node, field_type)

        elif isinstance(node_value, ast.VariableNode):
            input_field = input_type.fields.get(node_value.name.value)
            if input_field is None:
                yield None
            else:
                value_node = ast_from_value(self.context.variables[node_value.name.value], input_field.type)
                yield ast.ObjectFieldNode(name=node_value.name, value=value_node)

        else:
            yield node



# Generated by Django 5.2 on 2025-06-01 12:26
from __future__ import annotations

from django.db import migrations, models

from undine.persisted_documents.validators import validate_document, validate_document_id


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="PersistedDocument",
            fields=[
                (
                    "document_id",
                    models.CharField(
                        max_length=255,
                        primary_key=True,
                        serialize=False,
                        validators=[validate_document_id],
                    ),
                ),
                ("document", models.TextField(validators=[validate_document])),
                ("created_at", models.DateTimeField(auto_now_add=True)),
            ],
            options={
                "verbose_name": "persisted document",
                "verbose_name_plural": "persisted documents",
                "abstract": False,
                "swappable": "UNDINE_PERSISTED_DOCUMENTS_MODEL",
            },
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.http import StreamingHttpResponse
from graphql import GraphQLError

from undine.exceptions import GraphQLErrorGroup
from undine.execution import execute_graphql_http_async, execute_graphql_http_sync
from undine.http.utils import (
    HttpEventSourcingNotAllowedResponse,
    get_http_version,
    graphql_result_response,
    require_graphql_request_async,
    require_graphql_request_sync,
)
from undine.parsers import GraphQLRequestParamsParser
from undine.settings import undine_settings
from undine.utils.graphql.server_sent_events import execute_graphql_sse_dc, result_to_sse_dc
from undine.utils.graphql.utils import get_error_execution_result

if TYPE_CHECKING:
    from collections.abc import AsyncIterable

    from undine.typing import DjangoRequestProtocol, DjangoResponseProtocol

__all__ = [
    "graphql_view_async",
    "graphql_view_sync",
]


@require_graphql_request_sync
def graphql_view_sync(request: DjangoRequestProtocol) -> DjangoResponseProtocol:
    """A sync view for GraphQL requests."""
    try:
        params = GraphQLRequestParamsParser.run(request)
    except (GraphQLError, GraphQLErrorGroup) as error:
        result = get_error_execution_result(error)
    else:
        result = execute_graphql_http_sync(params, request)

    return graphql_result_response(result, content_type=request.response_content_type)


@require_graphql_request_async
async def graphql_view_async(request: DjangoRequestProtocol) -> DjangoResponseProtocol:
    """An async view for GraphQL requests."""
    if request.response_content_type == "text/event-stream":
        # TODO: Single connection mode not implemented
        if get_http_version(request) < (2, 0) and not undine_settings.USE_SSE_DISTINCT_CONNECTIONS_FOR_HTTP_1:
            return HttpEventSourcingNotAllowedResponse()

        try:
            params = GraphQLRequestParamsParser.run(request)
        except (GraphQLError, GraphQLErrorGroup) as error:
            result = get_error_execution_result(error)
            event_stream = result_to_sse_dc(result)
        else:
            event_stream = execute_graphql_sse_dc(params=params, request=request)

        stream: AsyncIterable[str] = (event.encode() async for event in event_stream)
        return StreamingHttpResponse(stream, content_type=request.response_content_type)

    try:
        params = GraphQLRequestParamsParser.run(request)
    except (GraphQLError, GraphQLErrorGroup) as error:
        result = get_error_execution_result(error)
    else:
        result = await execute_graphql_http_async(params, request)

    return graphql_result_response(result, content_type=request.response_content_type)



from __future__ import annotations

from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator
from contextlib import aclosing, nullcontext, suppress
from functools import wraps
from inspect import isawaitable
from typing import TYPE_CHECKING, Any

from django.core.exceptions import ValidationError
from graphql import (
    ExecutionContext,
    ExecutionResult,
    GraphQLError,
    MiddlewareManager,
    OperationType,
    ParallelVisitor,
    TypeInfo,
    TypeInfoVisitor,
    ValidationContext,
    ast_from_value,
    is_non_null_type,
    parse,
    version_info,
    visit,
)

from undine.exceptions import (
    GraphQLAsyncNotSupportedError,
    GraphQLCannotUseHTTPForMutationsNonPostRequestError,
    GraphQLCannotUseHTTPForSubscriptionsError,
    GraphQLCannotUseSSEForMutationsError,
    GraphQLCannotUseSSEForMutationsNonPostRequestError,
    GraphQLCannotUseSSEForQueriesError,
    GraphQLCannotUseWebSocketsForMutationsError,
    GraphQLCannotUseWebSocketsForQueriesError,
    GraphQLErrorGroup,
    GraphQLNoExecutionResultError,
    GraphQLSubscriptionNoEventStreamError,
    GraphQLUnexpectedError,
    GraphQLValidationAbortedError,
)
from undine.hooks import (
    ExecutionLifecycleHookManager,
    LifecycleHook,
    LifecycleHookContext,
    with_execution_lifecycle_hooks_manager,
    with_execution_lifecycle_hooks_manager_async,
    with_operation_lifecycle_hooks_manager,
    with_operation_lifecycle_hooks_manager_async,
    with_parse_lifecycle_hooks_manager,
    with_parse_lifecycle_hooks_manager_async,
    with_validation_lifecycle_hooks_manager,
    with_validation_lifecycle_hooks_manager_async,
)
from undine.http.utils import is_sse_request, is_websocket_request
from undine.settings import undine_settings
from undine.utils.graphql.utils import (
    get_error_execution_result,
    get_operation_type,
    graphql_errors_hook,
    located_validation_error,
)
from undine.utils.graphql.validation_rules import get_validation_rules
from undine.utils.reflection import cancel_awaitable

try:
    # graphql-core >= 3.3.0
    from graphql.execution.execute import execute_subscription
except ImportError:
    from graphql.execution.subscribe import execute_subscription


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from graphql import DocumentNode, GraphQLInputType, GraphQLOutputType, GraphQLSchema, Node, ValueNode
    from graphql.execution.build_field_plan import FieldGroup
    from graphql.execution.execute import IncrementalContext
    from graphql.pyutils import AwaitableOrValue, Path

    from undine.dataclasses import GraphQLHttpParams
    from undine.typing import DjangoRequestProtocol, ExecutionResultGen, P, SubscriptionResult


__all__ = [
    "execute_graphql_http_async",
    "execute_graphql_http_sync",
    "execute_graphql_with_subscription",
]


# HTTP sync execution


def raised_exceptions_as_execution_results_sync(
    func: Callable[P, ExecutionResult],
) -> Callable[P, ExecutionResult]:
    """Wraps raised exceptions as GraphQL ExecutionResults if they happen in `execute_graphql_sync`."""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> ExecutionResult:
        try:
            return func(*args, **kwargs)

        except GraphQLError as error:
            return get_error_execution_result(error)

        except GraphQLErrorGroup as error:
            return get_error_execution_result(error)

        except Exception as error:  # noqa: BLE001
            return get_error_execution_result(GraphQLUnexpectedError(message=str(error)))

    return wrapper


@raised_exceptions_as_execution_results_sync
def execute_graphql_http_sync(params: GraphQLHttpParams, request: DjangoRequestProtocol) -> ExecutionResult:
    """
    Executes a GraphQL operation received from an HTTP request synchronously.
    Assumes that the schema has been validated (e.g. created using `undine.schema.create_schema`).

    :param params: GraphQL request parameters.
    :param request: The request to use as the GraphQL execution context value.
    """
    context = LifecycleHookContext.from_graphql_params(params=params, request=request)
    return _run_operation_sync(context)


@with_operation_lifecycle_hooks_manager
def _run_operation_sync(context: LifecycleHookContext) -> ExecutionResult:
    if context.result is None:
        _parse_source_sync(context)
        if context.result is None:
            _validate_document_sync(context)
            if context.result is None:
                return _execute_sync(context)

    if isawaitable(context.result):
        cancel_awaitable(context.result)
        context.result = get_error_execution_result(GraphQLAsyncNotSupportedError())
        return context.result

    return context.result


@with_parse_lifecycle_hooks_manager
def _parse_source_sync(context: LifecycleHookContext) -> None:
    if context.result is not None:
        return

    if context.document is not None:
        return

    try:
        context.document = parse(
            source=context.source,
            no_location=undine_settings.NO_ERROR_LOCATION,
            max_tokens=undine_settings.MAX_TOKENS,
        )
    except GraphQLError as error:
        context.result = get_error_execution_result(error)


@with_validation_lifecycle_hooks_manager
def _validate_document_sync(context: LifecycleHookContext) -> None:
    if context.result is not None:
        return

    validation_errors = _validate(
        document=context.document,
        variables=context.variables,
        request=context.request,
    )
    if validation_errors:
        context.result = get_error_execution_result(validation_errors)
        return

    _validate_http(context)
    if context.result is not None:
        return


@with_execution_lifecycle_hooks_manager
def _execute_sync(context: LifecycleHookContext) -> ExecutionResult:
    try:
        exec_context = _get_execution_context(
            document=context.document,  # type: ignore[arg-type]
            root_value=undine_settings.ROOT_VALUE,
            context_value=context.request,
            variable_values=context.variables,
            operation_name=context.operation_name,
            middleware=_get_middleware_manager(context.lifecycle_hooks),
        )
    except GraphQLErrorGroup as error:
        context.result = get_error_execution_result(error)
        return context.result

    result = _execute(exec_context)

    if exec_context.is_awaitable(result):
        cancel_awaitable(result)
        context.result = get_error_execution_result(GraphQLAsyncNotSupportedError())
        return context.result

    if version_info >= (3, 3, 0):
        from graphql import ExperimentalIncrementalExecutionResults  # noqa: PLC0415
        from graphql.execution.execute import UNEXPECTED_MULTIPLE_PAYLOADS  # noqa: PLC0415

        if isinstance(result, ExperimentalIncrementalExecutionResults):
            context.result = get_error_execution_result(GraphQLError(UNEXPECTED_MULTIPLE_PAYLOADS))
            return context.result

    context.result = result
    return context.result


# HTTP async execution


def raised_exceptions_as_execution_results_async(
    func: Callable[P, Awaitable[ExecutionResult]],
) -> Callable[P, Awaitable[ExecutionResult]]:
    """Wraps raised exceptions as GraphQL ExecutionResults if they happen in `execute_graphql_async`."""

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> ExecutionResult:
        try:
            return await func(*args, **kwargs)

        except GraphQLError as error:
            return get_error_execution_result(error)

        except GraphQLErrorGroup as error:
            return get_error_execution_result(error)

        except Exception as error:  # noqa: BLE001
            return get_error_execution_result(GraphQLUnexpectedError(message=str(error)))

    return wrapper


@raised_exceptions_as_execution_results_async
async def execute_graphql_http_async(params: GraphQLHttpParams, request: DjangoRequestProtocol) -> ExecutionResult:
    """
    Executes a GraphQL operation received from an HTTP request asynchronously.
    Assumes that the schema has been validated (e.g. created using `undine.schema.create_schema`).

    :param params: GraphQL request parameters.
    :param request: The request to use as the GraphQL execution context value.
    """
    context = LifecycleHookContext.from_graphql_params(params=params, request=request)
    return await _run_operation_async(context)


@with_operation_lifecycle_hooks_manager_async
async def _run_operation_async(context: LifecycleHookContext) -> ExecutionResult:
    if context.result is None:
        await _parse_source_async(context)
        if context.result is None:
            await _validate_document_async(context)
            if context.result is None:
                return await _execute_async(context)

    if isinstance(context.result, AsyncIterator):
        context.result = get_error_execution_result(GraphQLCannotUseHTTPForSubscriptionsError())
        return context.result

    if isawaitable(context.result):
        context.result = await context.result  # type: ignore[assignment]

    if version_info >= (3, 3, 0):
        from graphql import ExperimentalIncrementalExecutionResults  # noqa: PLC0415
        from graphql.execution.execute import UNEXPECTED_MULTIPLE_PAYLOADS  # noqa: PLC0415

        if isinstance(context.result, ExperimentalIncrementalExecutionResults):
            context.result = get_error_execution_result(GraphQLError(UNEXPECTED_MULTIPLE_PAYLOADS))
            return context.result

    return context.result  # type: ignore[return-value]


@with_parse_lifecycle_hooks_manager_async
async def _parse_source_async(context: LifecycleHookContext) -> None:  # noqa: RUF029
    if context.result is not None:
        return

    if context.document is not None:
        return

    try:
        context.document = parse(
            source=context.source,
            no_location=undine_settings.NO_ERROR_LOCATION,
            max_tokens=undine_settings.MAX_TOKENS,
        )
    except GraphQLError as error:
        context.result = get_error_execution_result(error)


@with_validation_lifecycle_hooks_manager_async
async def _validate_document_async(context: LifecycleHookContext) -> None:  # noqa: RUF029
    if context.result is not None:
        return

    validation_errors = _validate(
        document=context.document,
        variables=context.variables,
        request=context.request,
    )
    if validation_errors:
        context.result = get_error_execution_result(validation_errors)
        return

    if is_websocket_request(context.request):
        _validate_websockets(context)
    elif is_sse_request(context.request):
        _validate_sse(context)
    else:
        _validate_http(context)


@with_execution_lifecycle_hooks_manager_async
async def _execute_async(context: LifecycleHookContext) -> ExecutionResult:
    try:
        exec_context = _get_execution_context(
            document=context.document,  # type: ignore[arg-type]
            root_value=undine_settings.ROOT_VALUE,
            context_value=context.request,
            variable_values=context.variables,
            operation_name=context.operation_name,
            middleware=_get_middleware_manager(context.lifecycle_hooks),
        )
    except GraphQLErrorGroup as error:
        context.result = get_error_execution_result(error)
        return context.result

    result = _execute(exec_context)

    if result is None:  # pragma: no cover
        context.result = get_error_execution_result(GraphQLNoExecutionResultError())
        return context.result

    context.result = await result if exec_context.is_awaitable(result) else result

    if version_info >= (3, 3, 0):
        from graphql import ExperimentalIncrementalExecutionResults  # noqa: PLC0415
        from graphql.execution.execute import UNEXPECTED_MULTIPLE_PAYLOADS  # noqa: PLC0415

        if isinstance(context.result, ExperimentalIncrementalExecutionResults):
            context.result = get_error_execution_result(GraphQLError(UNEXPECTED_MULTIPLE_PAYLOADS))
            return context.result

    return context.result


# Subscription enabled execution


def raised_exceptions_as_execution_results_with_subscriptions(
    func: Callable[P, Awaitable[SubscriptionResult]],
) -> Callable[P, Awaitable[SubscriptionResult]]:
    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> SubscriptionResult:
        try:
            return await func(*args, **kwargs)

        except GraphQLError as error:
            return get_error_execution_result(error)

        except GraphQLErrorGroup as error:
            return get_error_execution_result(error)

        except Exception as error:  # noqa: BLE001
            return get_error_execution_result(GraphQLUnexpectedError(message=str(error)))

    return wrapper


@raised_exceptions_as_execution_results_with_subscriptions
async def execute_graphql_with_subscription(
    params: GraphQLHttpParams,
    request: DjangoRequestProtocol,
) -> SubscriptionResult:
    """
    Executes a GraphQL operation with subscriptions support asynchronously.
    Assumes that the schema has been validated (e.g. created using `undine.schema.create_schema`).

    :param params: GraphQL request parameters.
    :param request: The request to use as the GraphQL execution context value.
    """
    context = LifecycleHookContext.from_graphql_params(params=params, request=request)
    return await _run_operation_with_subscription(context)


@with_operation_lifecycle_hooks_manager_async
async def _run_operation_with_subscription(context: LifecycleHookContext) -> SubscriptionResult:
    if context.result is None:
        await _parse_source_async(context)
        if context.result is None:
            await _validate_document_async(context)
            if context.result is None:
                operation_type = get_operation_type(context.document, context.operation_name)
                if operation_type == OperationType.SUBSCRIPTION:
                    return await _subscribe(context)
                return await _execute_async(context)

    if isawaitable(context.result):
        context.result = await context.result  # type: ignore[assignment]

    return context.result  # type: ignore[return-value]


async def _subscribe(context: LifecycleHookContext) -> SubscriptionResult:
    """Executes a subscription operation. See: `graphql.execution.subscribe.subscribe`."""
    result_or_stream = await _create_source_event_stream(context=context)
    if isinstance(result_or_stream, ExecutionResult):
        return result_or_stream

    return _map_source_to_response(source=result_or_stream, context=context)


async def _create_source_event_stream(context: LifecycleHookContext) -> AsyncIterable[Any] | ExecutionResult:
    """
    A source event stream represents a sequence of events,
    each of which triggers a GraphQL execution for that event.
    """
    try:
        exec_context = _get_execution_context(
            document=context.document,  # type: ignore[arg-type]
            root_value=undine_settings.ROOT_VALUE,
            context_value=context.request,
            variable_values=context.variables,
            operation_name=context.operation_name,
            middleware=_get_middleware_manager(context.lifecycle_hooks),
        )
    except GraphQLErrorGroup as error:
        return get_error_execution_result(error)

    try:
        event_stream = execute_subscription(exec_context)
        if exec_context.is_awaitable(event_stream):
            event_stream = await event_stream

    except GraphQLError as error:
        return get_error_execution_result(error)

    except GraphQLErrorGroup as error:
        return get_error_execution_result(error)

    if not isinstance(event_stream, AsyncIterable):
        return get_error_execution_result(GraphQLSubscriptionNoEventStreamError())

    return event_stream


async def _map_source_to_response(source: AsyncIterable, context: LifecycleHookContext) -> ExecutionResultGen:
    """
    For each payload yielded from a subscription,
    map it over the normal GraphQL `execute` function, with `payload` as the `root_value`.
    """
    manager = aclosing(source) if isinstance(source, AsyncGenerator) else nullcontext()

    async with manager:
        stream = aiter(source)

        while True:
            context.result = None

            async with ExecutionLifecycleHookManager(hooks=context.lifecycle_hooks):
                if context.result is not None:
                    yield context.result
                    continue

                try:
                    payload = await anext(stream)
                except StopAsyncIteration:
                    break

                if isinstance(payload, GraphQLError):
                    context.result = get_error_execution_result(payload)
                    yield context.result
                    continue

                if isinstance(payload, GraphQLErrorGroup):
                    context.result = get_error_execution_result(payload)
                    yield context.result
                    continue

                exec_context = _get_execution_context(
                    document=context.document,
                    root_value=payload,
                    context_value=context.request,
                    variable_values=context.variables,
                    operation_name=context.operation_name,
                    middleware=_get_middleware_manager(context.lifecycle_hooks),
                )
                result = _execute(exec_context)

                context.result = await result if exec_context.is_awaitable(result) else result
                yield context.result


# Helpers


def _validate_http(context: LifecycleHookContext) -> None:
    operation_type = get_operation_type(context.document, context.operation_name)

    if operation_type == OperationType.MUTATION and context.request.method != "POST":
        context.result = get_error_execution_result(GraphQLCannotUseHTTPForMutationsNonPostRequestError())
        return

    if operation_type == OperationType.SUBSCRIPTION:
        context.result = get_error_execution_result(GraphQLCannotUseHTTPForSubscriptionsError())
        return


def _validate_sse(context: LifecycleHookContext) -> None:
    operation_type = get_operation_type(context.document, context.operation_name)

    if operation_type == OperationType.QUERY and not undine_settings.ALLOW_QUERIES_WITH_SSE:
        context.result = get_error_execution_result(GraphQLCannotUseSSEForQueriesError())
        return

    if operation_type == OperationType.MUTATION:
        if not undine_settings.ALLOW_MUTATIONS_WITH_SSE:
            context.result = get_error_execution_result(GraphQLCannotUseSSEForMutationsError())
            return

        if context.request.method != "POST":
            context.result = get_error_execution_result(GraphQLCannotUseSSEForMutationsNonPostRequestError())
            return


def _validate_websockets(context: LifecycleHookContext) -> None:
    operation_type = get_operation_type(context.document, context.operation_name)

    if operation_type == OperationType.QUERY and not undine_settings.ALLOW_QUERIES_WITH_WEBSOCKETS:
        context.result = get_error_execution_result(GraphQLCannotUseWebSocketsForQueriesError())
        return

    if operation_type == OperationType.MUTATION and not undine_settings.ALLOW_MUTATIONS_WITH_WEBSOCKETS:
        context.result = get_error_execution_result(GraphQLCannotUseWebSocketsForMutationsError())
        return


def _get_execution_context(
    *,
    document: DocumentNode,
    root_value: Any,
    context_value: Any,
    variable_values: dict[str, Any],
    operation_name: str | None,
    middleware: MiddlewareManager | None,
) -> UndineExecutionContext:
    context_or_errors = undine_settings.EXECUTION_CONTEXT_CLASS.build(
        schema=undine_settings.SCHEMA,
        document=document,
        root_value=root_value,
        context_value=context_value,
        raw_variable_values=variable_values,
        operation_name=operation_name,
        middleware=middleware,
    )

    if isinstance(context_or_errors, list):
        raise GraphQLErrorGroup(errors=context_or_errors)

    return context_or_errors


def _get_middleware_manager(lifecycle_hooks: list[LifecycleHook]) -> MiddlewareManager | None:
    hooks = [hook for hook in lifecycle_hooks if hook.__class__.resolve != LifecycleHook.resolve]
    return MiddlewareManager(*reversed(hooks)) if hooks else None


def _validate(
    document: DocumentNode,
    variables: dict[str, Any] | None = None,
    request: DjangoRequestProtocol | None = None,
) -> list[GraphQLError]:
    errors: list[GraphQLError] = []

    def on_error(error: GraphQLError) -> None:
        if len(errors) >= undine_settings.MAX_ERRORS:
            errors.append(GraphQLValidationAbortedError())
            raise GraphQLValidationAbortedError

        errors.append(error)

    type_info = TypeInfo(schema=undine_settings.SCHEMA)

    context = UndineValidationContext(
        schema=undine_settings.SCHEMA,
        document=document,
        variables=variables or {},
        request=request,
        type_info=type_info,
        on_error=on_error,
    )

    visitors = [rule(context) for rule in get_validation_rules(inside_request=request is not None)]

    with suppress(GraphQLValidationAbortedError):
        visit(document, TypeInfoVisitor(type_info, ParallelVisitor(visitors)))

    return errors


def _execute(context: UndineExecutionContext) -> ExecutionResult:
    if version_info < (3, 3, 0):
        return _execute_old(context)
    return _execute_new(context)


def _execute_old(context: UndineExecutionContext) -> AwaitableOrValue[ExecutionResult]:
    """Execution for graphql-core < 3.3.0."""
    try:
        data_or_awaitable = context.execute_operation(context.operation, context.root_value)

    except GraphQLError as error:
        context.errors.append(error)
        return get_error_execution_result(context.errors)

    except GraphQLErrorGroup as error:
        context.errors.extend(error.flatten())
        return get_error_execution_result(context.errors)

    if context.is_awaitable(data_or_awaitable):

        async def await_result() -> ExecutionResult:
            try:
                data = await data_or_awaitable

            except GraphQLError as err:
                context.errors.append(err)
                return get_error_execution_result(context.errors)

            except GraphQLErrorGroup as err:
                context.errors.extend(err.flatten())
                return get_error_execution_result(context.errors)

            else:
                graphql_errors_hook(context.errors)
                return ExecutionResult(data=data, errors=context.errors or None)

        return await_result()

    graphql_errors_hook(context.errors)
    return ExecutionResult(data=data_or_awaitable, errors=context.errors or None)


def _execute_new(context: UndineExecutionContext) -> AwaitableOrValue[ExecutionResult]:
    """Execution for graphql-core >= 3.3.0a10."""
    from graphql import ExperimentalIncrementalExecutionResults  # noqa: PLC0415

    try:
        data = context.execute_operation()

    except GraphQLError as error:
        context.errors = context.errors or []
        context.errors.append(error)
        return get_error_execution_result(context.errors)

    except GraphQLErrorGroup as err:
        context.errors = context.errors or []
        context.errors.extend(err.flatten())
        return get_error_execution_result(context.errors)

    if context.is_awaitable(data):

        async def await_result() -> ExecutionResult:
            try:
                awaited_data = await data

            except GraphQLError as error:
                context.errors = context.errors or []
                context.errors.append(error)
                return get_error_execution_result(context.errors)

            except GraphQLErrorGroup as err:
                context.errors = context.errors or []
                context.errors.extend(err.flatten())
                return get_error_execution_result(context.errors)

            if isinstance(awaited_data, ExecutionResult) and awaited_data.errors is not None:
                graphql_errors_hook(awaited_data.errors)

            if (
                isinstance(awaited_data, ExperimentalIncrementalExecutionResults)
                and awaited_data.initial_result.errors is not None
            ):
                graphql_errors_hook(awaited_data.initial_result.errors)

            return awaited_data

        return await_result()

    if isinstance(data, ExecutionResult) and data.errors is not None:
        graphql_errors_hook(data.errors)

    if isinstance(data, ExperimentalIncrementalExecutionResults) and data.initial_result.errors is not None:
        graphql_errors_hook(data.initial_result.errors)

    return data


# Contexts


class UndineExecutionContext(ExecutionContext):
    """Custom GraphQL execution context class."""

    if version_info >= (3, 3, 0):

        def handle_field_error(
            self,
            raw_error: Exception,
            return_type: GraphQLOutputType,
            field_group: FieldGroup,
            path: Path,
            incremental_context: IncrementalContext | None = None,
        ) -> None:
            from graphql.execution.execute import to_nodes  # noqa: PLC0415

            match raw_error:
                case ValidationError():
                    error_group = located_validation_error(raw_error, to_nodes(field_group), path.as_list())
                    self.handle_field_errors_group(error_group, return_type, field_group, path, incremental_context)

                case GraphQLErrorGroup():
                    self.handle_field_errors_group(raw_error, return_type, field_group, path, incremental_context)

                case _:
                    super().handle_field_error(
                        raw_error=raw_error,
                        return_type=return_type,
                        field_group=field_group,
                        path=path,
                        incremental_context=incremental_context,
                    )

        def handle_field_errors_group(
            self,
            raw_error: GraphQLErrorGroup,
            return_type: GraphQLOutputType,
            field_group: FieldGroup,
            path: Path,
            incremental_context: IncrementalContext | None = None,
        ) -> None:
            from graphql.execution.execute import to_nodes  # noqa: PLC0415

            for err in raw_error.flatten():
                if not err.path:
                    err.path = path.as_list()
                if not err.nodes:
                    err.nodes = to_nodes(field_group)

            if is_non_null_type(return_type):
                raise raw_error

            for err in raw_error.flatten():
                self.handle_field_error(
                    raw_error=err,
                    return_type=return_type,
                    field_group=field_group,
                    path=path,
                    incremental_context=incremental_context,
                )

    else:

        def handle_field_error(self, error: GraphQLError, return_type: GraphQLOutputType) -> None:
            raw_error: Exception = error.original_error or error
            path = error.path or []
            field_nodes = error.nodes or []

            match raw_error:
                case ValidationError():
                    error_group = located_validation_error(raw_error, field_nodes, path)
                    self.handle_field_errors_group(error_group, return_type, field_nodes, path)

                case GraphQLErrorGroup():
                    self.handle_field_errors_group(raw_error, return_type, field_nodes, path)

                case _:
                    super().handle_field_error(error=error, return_type=return_type)

        def handle_field_errors_group(
            self,
            raw_error: GraphQLErrorGroup,
            return_type: GraphQLOutputType,
            field_nodes: list[Node],
            path: list[str | int],
        ) -> None:
            for err in raw_error.flatten():
                if not err.path:
                    err.path = path
                if not err.nodes:
                    err.nodes = field_nodes

            if is_non_null_type(return_type):
                raise raw_error

            for err in raw_error.flatten():
                self.handle_field_error(err, return_type)


class UndineValidationContext(ValidationContext):
    """Custom GraphQL validation context class."""

    def __init__(  # noqa: PLR0917
        self,
        schema: GraphQLSchema,
        document: DocumentNode,
        variables: dict[str, Any],
        request: DjangoRequestProtocol | None,
        type_info: TypeInfo,
        on_error: Callable[[GraphQLError], None],
    ) -> None:
        super().__init__(schema=schema, ast=document, type_info=type_info, on_error=on_error)
        self.variables = variables
        self.request = request

    def variable_as_ast(self, variable: str, type_: GraphQLInputType) -> ValueNode | None:
        """Get the AST for the given variable."""
        value = self.variables.get(variable)
        if value is None:
            return None

        try:
            return ast_from_value(value, type_)
        except Exception:  # noqa: BLE001
            return None



"""Django settings for Undine. Can be configured in the Django settings file with the key 'UNDINE'."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from django.test.signals import setting_changed
from graphql import GraphQLField, GraphQLObjectType, GraphQLSchema, GraphQLString
from settings_holder import SettingsHolder, reload_settings

if TYPE_CHECKING:
    from collections.abc import Container

    from graphql import ASTValidationRule

    from undine.execution import UndineExecutionContext
    from undine.hooks import LifecycleHook
    from undine.optimizer.optimizer import QueryOptimizer
    from undine.typing import (
        DocstringParserProtocol,
        PersistedDocumentsPermissionsCallback,
        WebSocketConnectionInitHook,
        WebSocketConnectionPingHook,
        WebSocketConnectionPongHook,
    )
    from undine.utils.graphql.sdl_printer import SDLPrinter


__all__ = [
    "undine_settings",
]


SETTING_NAME: str = "UNDINE"


class UndineDefaultSettings(NamedTuple):
    """Default settings for Undine."""

    # Schema

    SCHEMA: GraphQLSchema = "undine.settings.example_schema"  # type: ignore[assignment]
    """The schema to use for the GraphQL API."""

    GRAPHQL_PATH: str = "graphql/"
    """The path where the GraphQL endpoint is located by default."""

    GRAPHQL_VIEW_NAME: str = "graphql"
    """The name of given to the GraphQL view in the URLconf."""

    # Flags

    AUTOGENERATION: bool = False
    """Whether to automatically generate fields & inputs for some Undine types or not."""

    ALLOW_DID_YOU_MEAN_SUGGESTIONS: bool = False
    """Whether to allow the 'did you mean' suggestions on error messages."""

    ALLOW_INTROSPECTION_QUERIES: bool = False
    """Whether schema introspection queries are allowed or not."""

    ASYNC: bool = False
    """Whether to use an async view for the GraphQL endpoint or not."""

    CAMEL_CASE_SCHEMA_FIELDS: bool = True
    """Should names be converted from 'snake_case' to 'camelCase' for the GraphQL schema?"""

    ENABLE_CLASS_ATTRIBUTE_DOCSTRINGS: bool = False
    """Whether to parse class attribute docstrings or not."""

    FILE_UPLOAD_ENABLED: bool = False
    """Whether file uploads are enabled. Should enable CSRF protection on the GraphiQL endpoint if enabled."""

    INCLUDE_ERROR_TRACEBACK: bool = False
    """Whether to include the error traceback in the response error extensions."""

    MUTATION_FULL_CLEAN: bool = True
    """Whether to run `model.full_clean()` when mutating a model."""

    EXPERIMENTAL_VISIBILITY_CHECKS: bool = False  # Experimental, may not work as expected
    """Whether to enable experimental visibility checks."""

    # Limits

    LIST_ENTRYPOINT_LIMIT: int | None = None
    """
    Maximum number of objects that can be returned from a list Entrypoint when not using pagination.
    If None, all items are fetched.
    """

    MAX_FILTERS_PER_TYPE: int = 20
    """The maximum number of filters allowed for a single `FilterSet`."""

    MAX_ORDERS_PER_TYPE: int = 10
    """The maximum number of orders allowed for a single `OrderSet`."""

    # Pagination

    PAGINATION_PAGE_SIZE: int | None = 100
    """The maximum number of items to return in a page when paginating."""

    PAGINATION_START_INDEX_KEY: str = "_undine_pagination_start"
    """The key to which the connection's pagination start index is annotated to or added to in the queryset hints."""

    PAGINATION_STOP_INDEX_KEY: str = "_undine_pagination_stop"
    """The key to which the connection's pagination stop index is annotated to or added to in the queryset hints."""

    PAGINATION_INDEX_KEY: str = "_undine_pagination_index"
    """The key to which nested connection node's pagination index is annotated to the queryset."""

    PAGINATION_TOTAL_COUNT_KEY: str = "_undine_pagination_total_count"
    """The key to which the connection's total count annotated to or added to in the queryset hints."""

    # GraphQL execution

    ADDITIONAL_VALIDATION_RULES: list[type[ASTValidationRule]] = []
    """Additional validation rules to use for validating the GraphQL schema."""

    EXECUTION_CONTEXT_CLASS: type[UndineExecutionContext] = "undine.execution.UndineExecutionContext"  # type: ignore[assignment]
    """GraphQL execution context class used by the schema."""

    LIFECYCLE_HOOKS: list[type[LifecycleHook]] = []
    """Lifecycle hooks to use during GraphQL operations."""

    MAX_ALLOWED_ALIASES: int = 15
    """The maximum number of aliases allowed in a single operation."""

    MAX_ALLOWED_DIRECTIVES: int = 50
    """The maximum number of directives allowed in a single operation."""

    MAX_ERRORS: int = 100
    """The maximum number of validation errors allowed in a GraphQL request before it is rejected."""

    MAX_QUERY_COMPLEXITY: int = 10
    """Maximum query complexity that is allowed to be queried in a single operation."""

    MAX_TOKENS: int | None = None
    """Maximum number of tokens the GraphQL parser will parse before it rejects a request"""

    MUTATION_INSTANCE_LIMIT: int = 100
    """The maximum number of objects that can be mutated in a single mutation."""

    NO_ERROR_LOCATION: bool = False
    """Whether to add the location information to GraphQL errors."""

    ROOT_VALUE: Any = None
    """The root value for the GraphQL execution."""

    # Testing client

    TESTING_CLIENT_FULL_STACKTRACE: bool = False
    """Whether to include the full stacktrace in testing client instead of just the relevant frames."""

    TESTING_CLIENT_NO_ASYNC_TIMEOUT: bool = False
    """Whether to disable the websocket timeouts in testing client."""

    # GraphiQL

    GRAPHIQL_ENABLED: bool = False
    """Is GraphiQL enabled?"""

    # Persisted documents

    PERSISTED_DOCUMENTS_ONLY: bool = False
    """Whether to only allow persisted documents to be executed."""

    PERSISTED_DOCUMENTS_PATH: str = "persisted-documents/"
    """The path where the persisted documents registration endpoint is located by default."""

    PERSISTED_DOCUMENTS_PERMISSION_CALLBACK: PersistedDocumentsPermissionsCallback = (
        "undine.persisted_documents.utils.default_permission_callback"  # type: ignore[assignment]
    )
    """The function to use for permission checks for registration of persisted documents."""

    PERSISTED_DOCUMENTS_VIEW_NAME: str = "persisted_documents"
    """The name of given to the persisted documents registration view in the URLconf."""

    # WebSocket

    ALLOW_QUERIES_WITH_WEBSOCKETS: bool = False
    """Whether queries can be executed over WebSockets."""

    ALLOW_MUTATIONS_WITH_WEBSOCKETS: bool = False
    """Whether mutations can be executed over WebSockets."""

    WEBSOCKET_CONNECTION_INIT_HOOK: WebSocketConnectionInitHook = "undine.utils.graphql.websocket.connection_init_hook"  # type: ignore[assignment]
    """The function to use for custom `ConnectionInit` logic."""

    WEBSOCKET_CONNECTION_INIT_TIMEOUT_SECONDS: int = 3
    """The number of seconds to wait for the `ConnectionInit` message after opening a WebSocket before closing it."""

    WEBSOCKET_PATH: str = "graphql/"
    """The path where the GraphQL over WebSocket endpoint is located."""

    WEBSOCKET_PING_HOOK: WebSocketConnectionPingHook = "undine.utils.graphql.websocket.ping_hook"  # type: ignore[assignment]
    """The function for specifying custom `Ping` message logic."""

    WEBSOCKET_PONG_HOOK: WebSocketConnectionPongHook = "undine.utils.graphql.websocket.pong_hook"  # type: ignore[assignment]
    """The function to for specifying custom `Pong` message logic."""

    # Server-Sent Events

    ALLOW_QUERIES_WITH_SSE: bool = False
    """Whether queries can be executed over Server-Sent Events."""

    ALLOW_MUTATIONS_WITH_SSE: bool = False
    """Whether mutations can be executed over Server-Sent Events."""

    USE_SSE_DISTINCT_CONNECTIONS_FOR_HTTP_1: bool = False
    """Whether Server-Sent Events should use distinct connections mode without a HTTP/2 connection."""

    # Django-modeltranslation

    MODELTRANSLATION_INCLUDE_TRANSLATABLE: bool = False
    """Whether to add translatable fields to the GraphQL schema when using `django-modeltranslation`."""

    MODELTRANSLATION_INCLUDE_TRANSLATIONS: bool = True
    """Whether to add translation fields to the GraphQL schema when using `django-modeltranslation`."""

    # Optimizer

    DISABLE_ONLY_FIELDS_OPTIMIZATION: bool = False
    """Disable optimizing fetched fields with `queryset.only()`."""

    OPTIMIZER_CLASS: type[QueryOptimizer] = "undine.optimizer.optimizer.QueryOptimizer"  # type: ignore[assignment]
    """The optimizer class to use for optimizing queries."""

    PREFETCH_HACK_CACHE_KEY: str = "_undine_prefetch_hack_cache"
    """The key to use for storing the prefetch hack cache in the queryset hints."""

    # Argument & parameter names

    MUTATION_INPUT_DATA_KEY: str = "input"
    """The key used for the input argument of a MutationType."""

    QUERY_TYPE_FILTER_INPUT_KEY: str = "filter"
    """The key used for the filter argument of QueryType."""

    QUERY_TYPE_ORDER_INPUT_KEY: str = "orderBy"
    """The key used for the order by argument of a `QueryType`."""

    RESOLVER_ROOT_PARAM_NAME: str = "root"
    """The name of the root/parent parameter in resolvers."""

    TOTAL_COUNT_PARAM_NAME: str = "totalCount"
    """The name of the total count parameter in connection resolvers."""

    # Other

    DOCSTRING_PARSER: type[DocstringParserProtocol] = "undine.parsers.parse_docstring.RSTDocstringParser"  # type: ignore[assignment]
    """The docstring parser to use."""

    SDL_PRINTER: type[SDLPrinter] = "undine.utils.graphql.sdl_printer.SDLPrinter"  # type: ignore[assignment]
    """The SDL printer to use."""

    PG_TEXT_SEARCH_PREFIX: str = "_undine_ts_vector"
    """A prefix to use for the filter aliases of postgres full text search Filters."""

    EMPTY_VALUES: Container[Any] = (None, "", [], {})
    """By default, if a Filter receives any of these values, it will be ignored."""

    # Extensions keys

    CALCULATION_ARGUMENT_EXTENSIONS_KEY: str = "undine_calculation_argument"
    """The key to use for storing the calculation in the extensions of the GraphQL type."""

    CONNECTION_EXTENSIONS_KEY: str = "undine_connection"
    """The key to use for storing the connection in the extensions of the GraphQL type."""

    OFFSET_PAGINATION_EXTENSIONS_KEY: str = "undine_offset_pagination"
    """The key to use for storing the offset pagination in the extensions of the GraphQL field."""

    DIRECTIVE_ARGUMENT_EXTENSIONS_KEY: str = "undine_directive_argument"
    """The key used to store a Directive argument in the GraphQL extensions."""

    DIRECTIVE_EXTENSIONS_KEY: str = "undine_directive"
    """The key used to store a Directive in the GraphQL extensions."""

    ENTRYPOINT_EXTENSIONS_KEY: str = "undine_entrypoint"
    """The key used to store an Entrypoint in the field GraphQL extensions."""

    FIELD_EXTENSIONS_KEY: str = "undine_field"
    """The key used to store a Field in the field GraphQL extensions."""

    FILTER_EXTENSIONS_KEY: str = "undine_filter"
    """The key used to store a `Filter` in the argument GraphQL extensions."""

    FILTERSET_EXTENSIONS_KEY: str = "undine_filterset"
    """The key used to store a FilterSet in the argument GraphQL extensions."""

    INPUT_EXTENSIONS_KEY: str = "undine_input"
    """The key used to store an `Input` in the argument GraphQL extensions."""

    INTERFACE_FIELD_EXTENSIONS_KEY: str = "undine_interface_field"
    """The key used to store an `InterfaceField` in the field GraphQL extensions."""

    INTERFACE_TYPE_EXTENSIONS_KEY: str = "undine_interface"
    """The key used to store a `InterfaceType` in the object type GraphQL extensions."""

    MUTATION_TYPE_EXTENSIONS_KEY: str = "undine_mutation_type"
    """The key used to store a `MutationType` in the argument GraphQL extensions."""

    ORDER_EXTENSIONS_KEY: str = "undine_order"
    """The key used to store an `Order` in the argument GraphQL extensions."""

    ORDERSET_EXTENSIONS_KEY: str = "undine_orderset"
    """The key used to store a `OrderSet` in the argument GraphQL extensions."""

    QUERY_TYPE_EXTENSIONS_KEY: str = "undine_query_type"
    """The key used to store a `QueryType` in the object type GraphQL extensions."""

    ROOT_TYPE_EXTENSIONS_KEY: str = "undine_root_type"
    """The key used to store a `RootType` in the object type GraphQL extensions."""

    SCALAR_EXTENSIONS_KEY: str = "undine_scalar"
    """The key used to store a `Scalar` in the scalar GraphQL extensions."""

    SCHEMA_DIRECTIVES_EXTENSIONS_KEY: str = "undine_schema_directives"
    """The key used to store the schema directives in the schema GraphQL extensions."""

    UNION_TYPE_EXTENSIONS_KEY: str = "undine_union_type"
    """The key used to store a `UnionType` in the argument GraphQL extensions."""


DEFAULTS: dict[str, Any] = UndineDefaultSettings()._asdict()

IMPORT_STRINGS: set[str | bytes] = {
    "ADDITIONAL_VALIDATION_RULES.0",
    "DOCSTRING_PARSER",
    "EXECUTION_CONTEXT_CLASS",
    "LIFECYCLE_HOOKS.0",
    "OPTIMIZER_CLASS",
    "PERSISTED_DOCUMENTS_PERMISSION_CALLBACK",
    "SCHEMA",
    "SDL_PRINTER",
    "WEBSOCKET_CONNECTION_INIT_HOOK",
    "WEBSOCKET_PING_HOOK",
    "WEBSOCKET_PONG_HOOK",
}


REMOVED_SETTINGS: dict[str, Any] = {
    "ENTRYPOINT_LIMIT_PER_MODEL": "LIST_ENTRYPOINT_LIMIT",
    "CONNECTION_PAGE_SIZE": "PAGINATION_PAGE_SIZE",
    "CONNECTION_START_INDEX_KEY": "PAGINATION_START_INDEX_KEY",
    "CONNECTION_STOP_INDEX_KEY": "PAGINATION_STOP_INDEX_KEY",
    "CONNECTION_INDEX_KEY": "PAGINATION_INDEX_KEY",
    "CONNECTION_TOTAL_COUNT_KEY": "PAGINATION_TOTAL_COUNT_KEY",
    "OPERATION_HOOKS": "LIFECYCLE_HOOKS",
    "PARSE_HOOKS": "LIFECYCLE_HOOKS",
    "VALIDATION_HOOKS": "LIFECYCLE_HOOKS",
    "EXECUTION_HOOKS": "LIFECYCLE_HOOKS",
    "MIDDLEWARE": "LIFECYCLE_HOOKS",
}

undine_settings: UndineDefaultSettings = SettingsHolder(  # type: ignore[assignment]
    setting_name=SETTING_NAME,
    defaults=DEFAULTS,
    import_strings=IMPORT_STRINGS,
    removed_settings=REMOVED_SETTINGS,
)

reload_my_settings = reload_settings(SETTING_NAME, undine_settings)  # type: ignore[arg-type]
setting_changed.connect(reload_my_settings)


# Placeholder schema
example_schema = GraphQLSchema(
    query=GraphQLObjectType(
        "Query",
        fields={
            "testing": GraphQLField(
                GraphQLString,
                resolve=lambda obj, info: "Hello World",  # noqa: ARG005
            ),
        },
    ),
)



from __future__ import annotations

from contextlib import suppress
from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import Q
from django.db.models.constants import LOOKUP_SEP

from undine import Filter
from undine.converters import convert_to_filter_resolver
from undine.dataclasses import UnionFilterRef
from undine.resolvers import FilterFunctionResolver, FilterModelFieldResolver, FilterQExpressionResolver
from undine.typing import CombinableExpression, GraphQLFilterResolver, ModelField


@convert_to_filter_resolver.register
def _(ref: FunctionType, **kwargs: Any) -> GraphQLFilterResolver:
    return FilterFunctionResolver(func=ref)


@convert_to_filter_resolver.register
def _(_: ModelField, **kwargs: Any) -> GraphQLFilterResolver:
    caller: Filter = kwargs["caller"]
    lookup = f"{caller.field_name}{LOOKUP_SEP}{caller.lookup}"
    return FilterModelFieldResolver(lookup=lookup)


@convert_to_filter_resolver.register
def _(_: UnionFilterRef, **kwargs: Any) -> GraphQLFilterResolver:
    caller: Filter = kwargs["caller"]
    lookup = f"{caller.field_name}{LOOKUP_SEP}{caller.lookup}"
    return FilterModelFieldResolver(lookup=lookup)


@convert_to_filter_resolver.register
def _(ref: Q, **kwargs: Any) -> GraphQLFilterResolver:
    return FilterQExpressionResolver(q_expression=ref)


@convert_to_filter_resolver.register
def _(_: CombinableExpression, **kwargs: Any) -> GraphQLFilterResolver:
    # The expression or subquery should be aliased in the queryset.
    caller: Filter = kwargs["caller"]
    lookup = f"{caller.field_name}{LOOKUP_SEP}{caller.lookup}"
    return FilterModelFieldResolver(lookup=lookup)


@convert_to_filter_resolver.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> GraphQLFilterResolver:
    caller: Filter = kwargs["caller"]
    lookup = f"{caller.field_name}{LOOKUP_SEP}{caller.lookup}"
    return FilterModelFieldResolver(lookup=lookup)


with suppress(ImportError):
    from undine.utils.full_text_search import PostgresFTS, PostgresFTSExpressionResolver

    @convert_to_filter_resolver.register
    def _(ref: PostgresFTS, **kwargs: Any) -> GraphQLFilterResolver:
        return PostgresFTSExpressionResolver(fts=ref)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, Unpack

from graphql import DirectiveLocation, GraphQLField, Undefined

from undine.converters import (
    convert_to_description,
    convert_to_field_complexity,
    convert_to_field_ref,
    convert_to_field_resolver,
    convert_to_graphql_argument_map,
    convert_to_graphql_type,
    is_field_nullable,
    is_many,
)
from undine.dataclasses import MaybeManyOrNonNull
from undine.exceptions import MissingModelGenericError
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.typing import TModel
from undine.utils.graphql.type_registry import get_or_create_graphql_object_type
from undine.utils.graphql.utils import check_directives
from undine.utils.model_utils import get_default_manager, get_model_fields_for_graphql, get_related_name
from undine.utils.reflection import FunctionEqualityWrapper, cache_signature_if_function, get_members, get_wrapped_func
from undine.utils.registy import Registry
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from collections.abc import Collection, Container

    from django.db.models import Model, QuerySet
    from graphql import GraphQLArgumentMap, GraphQLFieldResolver, GraphQLObjectType, GraphQLOutputType

    from undine import FilterSet, GQLInfo, InterfaceType, OrderSet
    from undine.directives import Directive
    from undine.optimizer.optimizer import OptimizationData
    from undine.typing import (
        DjangoRequestProtocol,
        FieldParams,
        FieldPermFunc,
        OptimizerFunc,
        QueryTypeParams,
        VisibilityFunc,
    )

__all__ = [
    "Field",
    "QueryType",
]


class QueryTypeMeta(type):
    """A metaclass that modifies how a `QueryType` is created."""

    # Set in '__new__'
    __model__: type[Model]
    __filterset__: type[FilterSet] | None
    __orderset__: type[OrderSet] | None
    __field_map__: dict[str, Field]
    __schema_name__: str
    __interfaces__: Collection[type[InterfaceType]]
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[QueryTypeParams],
    ) -> QueryTypeMeta:
        if _name == "QueryType":  # Early return for the `QueryType` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        try:
            model = QueryTypeMeta.__model__
            del QueryTypeMeta.__model__
        except AttributeError as error:
            raise MissingModelGenericError(name=_name, cls="QueryType") from error

        auto = kwargs.get("auto", undine_settings.AUTOGENERATION)
        exclude = set(kwargs.get("exclude", []))
        if auto:
            exclude |= set(_attrs)
            _attrs |= get_fields_for_model(model, exclude=exclude)

        interfaces = kwargs.get("interfaces", [])
        for interface in interfaces:
            for field_name, interface_field in interface.__field_map__.items():
                _attrs.setdefault(field_name, interface_field.as_undine_field())

        query_type = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `Field` names.
        query_type.__model__ = model

        query_type.__filterset__ = kwargs.get("filterset")
        if query_type.__filterset__ is not None:
            query_type.__filterset__.__add_to_query_type__(query_type)  # type: ignore[arg-type]

        query_type.__orderset__ = kwargs.get("orderset")
        if query_type.__orderset__ is not None:
            query_type.__orderset__.__add_to_query_type__(query_type)  # type: ignore[arg-type]

        query_type.__field_map__ = get_members(query_type, Field)
        query_type.__schema_name__ = kwargs.get("schema_name", _name)
        query_type.__interfaces__ = interfaces
        query_type.__directives__ = kwargs.get("directives", [])
        query_type.__extensions__ = kwargs.get("extensions", {})
        query_type.__attribute_docstrings__ = parse_class_attribute_docstrings(query_type)

        check_directives(query_type.__directives__, location=DirectiveLocation.OBJECT)
        query_type.__extensions__[undine_settings.QUERY_TYPE_EXTENSIONS_KEY] = query_type

        for interface in interfaces:
            interface.__register_as_implementation__(query_type)  # type: ignore[arg-type]

        register = kwargs.get("register", True)
        if register:
            QUERY_TYPE_REGISTRY[model] = query_type  # type: ignore[assignment]

        for name, field in query_type.__field_map__.items():
            field.__connect__(query_type, name)  # type: ignore[arg-type]

        return query_type

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_object_type(cls.__output_type__())

    def __getitem__(cls, model: type[TModel]) -> type[QueryType[TModel]]:
        # Note that this should be cleaned up in '__new__',
        # but is not if an error occurs in the class body of the defined 'QueryType'!
        QueryTypeMeta.__model__ = model
        return cls  # type: ignore[return-value]

    def __output_type__(cls) -> GraphQLObjectType:
        """Creates a GraphQL `ObjectType` for this `QueryType`."""
        return get_or_create_graphql_object_type(
            name=cls.__schema_name__,
            fields=FunctionEqualityWrapper(cls.__output_fields__, context=cls),
            interfaces=[interface.__interface__() for interface in cls.__interfaces__],
            description=get_docstring(cls),
            is_type_of=cls.__is_type_of__,
            extensions=cls.__extensions__,
        )

    def __output_fields__(cls) -> dict[str, GraphQLField]:
        """Defer creating fields until all QueryTypes have been registered."""
        return {field.schema_name: field.as_graphql_field() for field in cls.__field_map__.values()}

    def __is_type_of__(cls, value: TModel, info: GQLInfo) -> bool:
        """
        Function for resolving types of abstract GraphQL types like unions.
        Indicates whether the given value belongs to this `QueryType`.
        """
        # Purposely not using `isinstance` here since models can inherit from other models.
        return type(value) is cls.__model__

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this query."""
        check_directives([directive], location=DirectiveLocation.OBJECT)
        cls.__directives__.append(directive)
        return cls


class QueryType(Generic[TModel], metaclass=QueryTypeMeta):
    """
    A class for creating a query in the GraphQL schema based on a Django Model.
    Represents a GraphQL `ObjectType` in the GraphQL schema.

    Must set the Django Model this `QueryType` is for using the Generic type parameter.

    The following parameters can be passed in the class definition:

    `filterset: type[FilterSet] = None`
        `FilterSet` class this `QueryType` uses for filtering.

    `orderset: type[OrderSet] = None`
        `OrderSet` class this `QueryType` uses for ordering.

    `auto: bool = <AUTOGENERATION setting>`
        Whether to add `Field` attributes for all Model fields automatically.

    `exclude: list[str] = []`
        Model fields to exclude from the automatically added `Field` attributes.

    `interfaces: list[type[InterfaceType]] = []`
        Interfaces this `QueryType` should implement.

    `register: bool = True`
        Whether to register the `QueryType` for the given Model.
        Only one `QueryType` can be registered per Model.
        Allows other `QueryTypes` to look up this `QueryType` for linking relations,
        and `MutationTypes` to find out their matching output type.

    `schema_name: str = <class name>`
        Override name for the `ObjectType` for the `QueryType` in the GraphQL schema.

    `directives: list[Directive] = []`
        `Directives` to add to the created `ObjectType`.

    `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `ObjectType`.

    >>> class TaskType(QueryType[Task]): ...
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `Field` names.

    # Set in metaclass
    __model__: ClassVar[type[Model]]
    __filterset__: ClassVar[type[FilterSet] | None]
    __orderset__: ClassVar[type[OrderSet] | None]
    __field_map__: ClassVar[dict[str, Field]]
    __schema_name__: ClassVar[str]
    __interfaces__: ClassVar[Collection[type[InterfaceType]]]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    @classmethod
    def __filter_queryset__(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        """Filtering that should always be applied when fetching objects through this QueryType."""
        return queryset

    @classmethod
    def __permissions__(cls, instance: TModel, info: GQLInfo) -> None:
        """Check permissions for accessing an instance through this `QueryType`."""

    @classmethod
    def __optimizations__(cls, data: OptimizationData, info: GQLInfo) -> None:
        """
        Hook for modifying the optimization data outside the GraphQL resolver context.
        Can be used to e.g. optimize data for permissions checks.
        """

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `QueryType` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True

    @classmethod
    def __get_queryset__(cls, info: GQLInfo) -> QuerySet[TModel]:
        """Base queryset for this `QueryType`."""
        return get_default_manager(cls.__model__).get_queryset()  # type: ignore[return-value]


class Field:
    """
    A class for defining a field for a `QueryType`.
    Represents a field on a GraphQL `ObjectType` for the `QueryType` this is added to.

    >>> class TaskType(QueryType[Task]):
    ...     name = Field()
    """

    def __init__(self, ref: Any = None, **kwargs: Unpack[FieldParams]) -> None:
        """
        Create a new Field.

        :param ref: Reference to build the `Field` from. Must be convertable by the `convert_to_field_ref` function.
                    If not provided, use the name of the attribute this is assigned to in the `QueryType` class.
        :param many: Whether the `Field` should return a non-null list of the referenced type.
        :param nullable: Whether the referenced type can be null.
        :param description: Description for the `Field`.
        :param deprecation_reason: If the `Field` is deprecated, describes the reason for deprecation.
        :param complexity: The complexity of resolving this field.
        :param field_name: Name of the field in the Django model. If not provided, use the name of the attribute.
        :param schema_name: Actual name of the `Field` in the GraphQL schema. Can be used to alias the `Field`
                            for the schema, or when the desired name is a Python keyword (e.g. `if` or `from`).
        :param directives: GraphQL directives for the `Field`.
        :param extensions: GraphQL extensions for the `Field`.
        """
        self.ref: Any = cache_signature_if_function(ref, depth=1)

        self.many: bool = kwargs.get("many", Undefined)  # type: ignore[assignment]
        self.nullable: bool = kwargs.get("nullable", Undefined)  # type: ignore[assignment]
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.complexity: int = kwargs.get("complexity", Undefined)  # type: ignore[assignment]
        self.field_name: str = kwargs.get("field_name", Undefined)  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.FIELD_DEFINITION)
        self.extensions[undine_settings.FIELD_EXTENSIONS_KEY] = self

        self.resolver_func: GraphQLFieldResolver | None = None
        self.optimizer_func: OptimizerFunc | None = None
        self.permissions_func: FieldPermFunc | None = None
        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, query_type: type[QueryType], name: str) -> None:
        """Connect this `Field` to the given `QueryType` using the given name."""
        self.query_type = query_type
        self.name = name
        self.field_name = self.field_name or name
        self.schema_name = self.schema_name or to_schema_name(name)

        if isinstance(self.ref, str):
            self.field_name = self.ref

        self.ref = convert_to_field_ref(self.ref, caller=self)

        if self.many is Undefined:
            self.many = is_many(self.ref, model=self.query_type.__model__, name=self.field_name)
        if self.nullable is Undefined:
            self.nullable = is_field_nullable(self.ref, caller=self)
        if self.complexity is Undefined:
            self.complexity = convert_to_field_complexity(self.ref, caller=self)
        if self.description is Undefined:
            self.description = self.query_type.__attribute_docstrings__.get(name)
            if self.description is None:
                self.description = convert_to_description(self.ref)

    def __call__(self, ref: GraphQLFieldResolver, /) -> Field:
        """Called when using as decorator with parenthesis: @Field(...)"""
        self.ref = cache_signature_if_function(ref, depth=1)
        return self

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.ref!r})>"

    def __str__(self) -> str:
        field = self.as_graphql_field()
        return undine_settings.SDL_PRINTER.print_field(self.schema_name, field, indent=False)

    def as_graphql_field(self) -> GraphQLField:
        return GraphQLField(
            type_=self.get_field_type(),
            args=self.get_field_arguments(),
            resolve=self.get_resolver(),
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            extensions=self.extensions,
        )

    def get_field_type(self) -> GraphQLOutputType:
        value = MaybeManyOrNonNull(self.ref, many=self.many, nullable=self.nullable)
        return convert_to_graphql_type(value, model=self.query_type.__model__)  # type: ignore[return-value]

    def get_field_arguments(self) -> GraphQLArgumentMap | None:
        return convert_to_graphql_argument_map(self.ref, many=self.many)

    def get_resolver(self) -> GraphQLFieldResolver:
        if self.resolver_func is not None:
            return convert_to_field_resolver(self.resolver_func, caller=self)
        return convert_to_field_resolver(self.ref, caller=self)

    def resolve(self, func: GraphQLFieldResolver | None = None, /) -> GraphQLFieldResolver:
        """
        Decorate a function to add a custom resolver for this Field.

        >>> class TaskType(QueryType[Task]):
        ...     name = Field()
        ...
        ...     @name.resolve
        ...     def resolve_name(self: Task, info: GQLInfo) -> str:
        ...         return self.name
        """
        if func is None:  # Allow `@<field_name>.resolve()`
            return self.resolve
        self.resolver_func = cache_signature_if_function(func, depth=1)
        return func

    def optimize(self, func: OptimizerFunc | None = None, /) -> OptimizerFunc:
        """
        Decorate a function to add custom optimization rules for this Field.

        >>> class TaskType(QueryType[Task]):
        ...     name = Field()
        ...
        ...     @name.optimize
        ...     def optimize_name(self: Field, data: OptimizationData, info: GQLInfo) -> None:
        ...         data.only_fields.add("name")
        """
        if func is None:  # Allow `@<field_name>.optimize()`
            return self.optimize  # type: ignore[return-value]
        self.optimizer_func = get_wrapped_func(func)
        return func

    def permissions(self, func: FieldPermFunc | None = None, /) -> FieldPermFunc:
        """
        Decorate a function to add it as a permission check for this Field.

        >>> class TaskType(QueryType[Task]):
        ...     name = Field()
        ...
        ...     @name.permissions
        ...     def name_permissions(self: Task, info: GQLInfo, value: str) -> None:
        ...         raise GraphQLPermissionError
        """
        if func is None:  # Allow `@<field_name>.permissions()`
            return self.permissions  # type: ignore[return-value]
        self.permissions_func = get_wrapped_func(func)
        return func

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the Field's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class TaskType(QueryType[Task]):
        ...     name = Field()
        ...
        ...     @name.visible
        ...     def name_visible(self: Field, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<field_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this field."""
        check_directives([directive], location=DirectiveLocation.FIELD_DEFINITION)
        self.directives.append(directive)
        return self


def get_fields_for_model(model: type[Model], *, exclude: Container[str] = ()) -> dict[str, Field]:
    """Add undine.Fields for all the given model's fields, except those in the 'exclude' list."""
    result: dict[str, Field] = {}

    for model_field in get_model_fields_for_graphql(model):
        field_name = get_related_name(model_field)  # type: ignore[arg-type]

        is_primary_key = bool(getattr(model_field, "primary_key", False))
        if is_primary_key:
            field_name = "pk"

        if field_name in exclude:
            continue

        result[field_name] = Field(model_field)

    return result


QUERY_TYPE_REGISTRY: Registry[type[Model], type[QueryType]] = Registry()
"""
Maps from a Django Model class to a corresponding `QueryType`.
This allows deferring the creation of field resolvers for related fields,
which would use a `QueryType` that is not created when the field is defined.
"""



from __future__ import annotations

import datetime
import uuid
from contextlib import suppress
from decimal import Decimal
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
from django.db.models import (
    BigIntegerField,
    BinaryField,
    BooleanField,
    CharField,
    DateField,
    DateTimeField,
    DecimalField,
    DurationField,
    FileField,
    FloatField,
    ImageField,
    IntegerField,
    JSONField,
    TextField,
    TimeField,
    UUIDField,
)
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute

from undine.converters import convert_model_field_to_python_type
from undine.typing import ToManyField, ToOneField
from undine.utils.model_fields import TextChoicesField


@convert_model_field_to_python_type.register
def _(_: CharField | TextField, **kwargs: Any) -> type:
    # CharField might have an enum, but we cannot access it anymore.
    return str


@convert_model_field_to_python_type.register
def _(ref: TextChoicesField, **kwargs: Any) -> type:
    return ref.choices_enum


@convert_model_field_to_python_type.register
def _(_: BooleanField, **kwargs: Any) -> type:
    return bool


@convert_model_field_to_python_type.register
def _(_: IntegerField | BigIntegerField, **kwargs: Any) -> type:
    return int


@convert_model_field_to_python_type.register
def _(_: FloatField, **kwargs: Any) -> type:
    return float


@convert_model_field_to_python_type.register
def _(_: DecimalField, **kwargs: Any) -> type:
    return Decimal


@convert_model_field_to_python_type.register
def _(_: DateField, **kwargs: Any) -> type:
    return datetime.date


@convert_model_field_to_python_type.register
def _(_: DateTimeField, **kwargs: Any) -> type:
    return datetime.datetime


@convert_model_field_to_python_type.register
def _(_: TimeField, **kwargs: Any) -> type:
    return datetime.time


@convert_model_field_to_python_type.register
def _(_: DurationField, **kwargs: Any) -> type:
    return datetime.timedelta


@convert_model_field_to_python_type.register
def _(_: BinaryField, **kwargs: Any) -> type:
    return bytes


@convert_model_field_to_python_type.register
def _(_: UUIDField, **kwargs: Any) -> type:
    return uuid.UUID


@convert_model_field_to_python_type.register
def _(_: FileField, **kwargs: Any) -> type:
    return str


@convert_model_field_to_python_type.register
def _(_: ImageField, **kwargs: Any) -> type:
    return str


@convert_model_field_to_python_type.register
def _(_: JSONField, **kwargs: Any) -> type:
    return dict[str, str]


@convert_model_field_to_python_type.register
def _(ref: ToManyField, **kwargs: Any) -> type:
    generic_type = convert_model_field_to_python_type(ref.target_field, **kwargs)
    return list.__class_getitem__(generic_type)


@convert_model_field_to_python_type.register
def _(ref: ToOneField, **kwargs: Any) -> type:
    return convert_model_field_to_python_type(ref.target_field, **kwargs)


@convert_model_field_to_python_type.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> type:
    return convert_model_field_to_python_type(ref.field, **kwargs)


@convert_model_field_to_python_type.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> type:
    return convert_model_field_to_python_type(ref.rel, **kwargs)


@convert_model_field_to_python_type.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> type:
    return convert_model_field_to_python_type(ref.related, **kwargs)


@convert_model_field_to_python_type.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> type:
    return convert_model_field_to_python_type(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_model_field_to_python_type.register
def _(ref: GenericRelation, **kwargs: Any) -> type:
    generic_type = convert_model_field_to_python_type(ref.target_field, **kwargs)
    return list.__class_getitem__(generic_type)


@convert_model_field_to_python_type.register
def _(ref: GenericRel, **kwargs: Any) -> type:
    return convert_model_field_to_python_type(ref.field)


@convert_model_field_to_python_type.register
def _(ref: GenericForeignKey, **kwargs: Any) -> type:
    field = ref.model._meta.get_field(ref.fk_field)
    return convert_model_field_to_python_type(field)


with suppress(ImportError):
    from django.contrib.postgres.fields import ArrayField, HStoreField

    @convert_model_field_to_python_type.register
    def _(ref: HStoreField, **kwargs: Any) -> type:
        return dict[str, str]

    @convert_model_field_to_python_type.register
    def _(ref: ArrayField, **kwargs: Any) -> type:
        item_type = convert_model_field_to_python_type(ref.base_field, **kwargs)
        return list.__class_getitem__(item_type)


with suppress(ImportError):
    from django.db.models import GeneratedField

    @convert_model_field_to_python_type.register
    def _(ref: GeneratedField, **kwargs: Any) -> type:
        return convert_model_field_to_python_type(ref.output_field, **kwargs)



from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING, Any, ClassVar, Self, Unpack

from graphql import DirectiveLocation, GraphQLArgument, Undefined

from undine.exceptions import (
    MissingDirectiveArgumentError,
    MissingDirectiveLocationsError,
    NotCompatibleWithError,
    UnexpectedDirectiveArgumentError,
)
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.utils.graphql.type_registry import get_or_create_graphql_directive
from undine.utils.graphql.utils import check_directives
from undine.utils.reflection import get_members, get_wrapped_func, has_callable_attribute
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from graphql import GraphQLDirective, GraphQLInputType

    from undine import CalculationArgument, Entrypoint, Field, Filter, Order
    from undine.entrypoint import RootTypeMeta
    from undine.filtering import FilterSetMeta
    from undine.interface import InterfaceTypeMeta
    from undine.mutation import MutationTypeMeta
    from undine.ordering import OrderSetMeta
    from undine.query import QueryTypeMeta
    from undine.typing import (
        DefaultValueType,
        DirectiveArgumentParams,
        DirectiveParams,
        DjangoRequestProtocol,
        T,
        VisibilityFunc,
    )
    from undine.union import UnionTypeMeta

__all__ = [
    "Directive",
    "DirectiveArgument",
]


class DirectiveMeta(type):
    """A metaclass that modifies how a `Directive` is created."""

    # Set in '__new__'
    __locations__: list[DirectiveLocation]
    __arguments__: dict[str, DirectiveArgument]
    __is_repeatable__: bool
    __schema_name__: str
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[DirectiveParams],
    ) -> DirectiveMeta:
        if _name == "Directive":  # Early return for the `Directive` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        locations = kwargs.get("locations", [])
        if locations is None:
            raise MissingDirectiveLocationsError(name=_name)

        directive = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `DirectiveArgument` names.
        directive.__locations__ = [DirectiveLocation(location) for location in locations]
        directive.__arguments__ = get_members(directive, DirectiveArgument)
        directive.__is_repeatable__ = kwargs.get("is_repeatable", False)
        directive.__schema_name__ = kwargs.get("schema_name", _name)
        directive.__extensions__ = kwargs.get("extensions", {})
        directive.__attribute_docstrings__ = parse_class_attribute_docstrings(directive)

        directive.__extensions__[undine_settings.DIRECTIVE_EXTENSIONS_KEY] = directive

        for name, argument in directive.__arguments__.items():
            argument.__connect__(directive, name)  # type: ignore[arg-type]

        # Create the GraphQL directive to register it.
        # This way it shows up in the GraphQL schema automatically.
        directive.__directive__()

        return directive

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_directive(cls.__directive__())

    def __directive__(cls) -> GraphQLDirective:
        """Creates the `GraphQLDirective` for this `Directive`."""
        return get_or_create_graphql_directive(
            name=cls.__schema_name__,
            locations=cls.__locations__,
            args={arg.schema_name: arg.as_graphql_argument() for arg in cls.__arguments__.values()},
            is_repeatable=cls.__is_repeatable__,
            description=get_docstring(cls),
            extensions=cls.__extensions__,
        )


class Directive(metaclass=DirectiveMeta):
    """
    A class for creating new Directives to add to GraphQL objects.
    Represents a GraphQL `Directive` in the `Schema`.

    The following parameters can be passed in the class definition:

    `locations: list[DirectiveLocation]`
        Places where this directive can be used. Required.

    `is_repeatable: bool = False`
        Whether the `Directive` is repeatable.

    `schema_name: str = <class name>`
        Override name for the `GraphQLDirective` for this `Directive` in the GraphQL schema.

    `directives`: `list[Directive] = []`
        `Directives` to add to the created `GraphQLDirective`.

    `extensions`: `dict[str, Any] = {}`
        GraphQL extensions for the created `GraphQLDirective`.

    >>> class MyDirective(Directive, locations=[DirectiveLocation.FIELD_DEFINITION]): ...
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `DirectiveArgument` names.

    # Set in metaclass
    __locations__: ClassVar[list[DirectiveLocation]]
    __arguments__: ClassVar[dict[str, DirectiveArgument]]
    __is_repeatable__: ClassVar[bool]
    __schema_name__: ClassVar[str]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    def __init__(self, **kwargs: Any) -> None:
        parameters: dict[str, Any] = {}

        for name, arg in self.__arguments__.items():
            value = kwargs.pop(name, arg.default_value)
            if value is Undefined:
                raise MissingDirectiveArgumentError(name=name, directive=type(self))

            parameters[name] = value

        if kwargs:
            raise UnexpectedDirectiveArgumentError(directive=type(self), kwargs=kwargs)

        self.__parameters__: MappingProxyType[str, Any] = MappingProxyType(parameters)

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `Directive` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True

    def __repr__(self) -> str:
        args = ", ".join(f"{name}={value!r}" for name, value in self.__parameters__.items())
        return f"<{dotpath(self.__class__)}({args})>"

    def __str__(self) -> str:
        return undine_settings.SDL_PRINTER.print_directive_usage(self, indent=False)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__parameters__ == other.__parameters__

    def __hash__(self) -> int:
        return hash((type(self), tuple(self.__parameters__.items())))

    def __call__(self, other: T, /) -> T:
        """
        Allow adding directives using decorators.

        >>> class MyDirective(Directive, locations=[DirectiveLocation.FIELD_DEFINITION]): ...
        >>>
        >>> @MyDirective()
        >>> class TaskType(QueryType[Task]): ...
        """
        self.__add_directive__(other)
        return other

    def __rmatmul__(self, other: T) -> T:
        """
        Allow adding directives using the @ operator.

        >>> class MyDirective(Directive, locations=[DirectiveLocation.FIELD_DEFINITION]): ...
        >>>
        >>> class TaskType(QueryType[Task]):
        >>>     name = Field() @ MyDirective()
        """
        self.__add_directive__(other)
        return other

    def __add_directive__(self, other: T) -> T:
        if has_callable_attribute(other, "add_directive"):
            other: CalculationArgument | Entrypoint | Field | Filter | Order
            other.add_directive(self)
            return other

        if has_callable_attribute(other, "__add_directive__"):
            other: (
                FilterSetMeta
                | InterfaceTypeMeta
                | MutationTypeMeta
                | OrderSetMeta
                | QueryTypeMeta
                | RootTypeMeta
                | UnionTypeMeta
            )
            other.__add_directive__(self)
            return other

        raise NotCompatibleWithError(obj=self, other=other)


class DirectiveArgument:
    """
    A class for defining a directive argument.
    Represents an argument on a GraphQL `Directive` for the `Directive` this is added to.

    >>> class MyDirective(Directive, locations=[DirectiveLocation.FIELD_DEFINITION]):
    ...     name = DirectiveArgument(GraphQLNonNull(GraphQLInt))
    """

    def __init__(self, input_type: GraphQLInputType, **kwargs: Unpack[DirectiveArgumentParams]) -> None:
        """
        Create a new `DirectiveArgument`.

        :param input_type: The input type to use for the `DirectiveArgument`.
        :param description: Description for the `DirectiveArgument`.
        :param default_value: Default value for the `DirectiveArgument`.
        :param deprecation_reason: If the `DirectiveArgument` is deprecated, describes the reason for deprecation.
        :param schema_name: Actual name in the GraphQL schema. Only needed if argument name is a python keyword.
        :param directives: GraphQL directives for the `DirectiveArgument`.
        :param extensions: GraphQL extensions for the `DirectiveArgument`.
        """
        self.input_type = input_type

        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.default_value: DefaultValueType = kwargs.get("default_value", Undefined)
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.ARGUMENT_DEFINITION)
        self.extensions[undine_settings.DIRECTIVE_ARGUMENT_EXTENSIONS_KEY] = self

        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, directive: type[Directive], name: str) -> None:
        """Connect this `DirectiveArgument` to the given `Directive` using the given name."""
        self.directive = directive
        self.name = name
        self.schema_name = self.schema_name or to_schema_name(name)

        if self.description is Undefined:
            self.description = self.directive.__attribute_docstrings__.get(name)

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(input_type={self.input_type!r})>"

    def __str__(self) -> str:
        arg = self.as_graphql_argument()
        return undine_settings.SDL_PRINTER.print_directive_argument(self.schema_name, arg, indent=False)

    def as_graphql_argument(self) -> GraphQLArgument:
        return GraphQLArgument(
            type_=self.input_type,
            default_value=self.default_value,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            out_name=self.name,
            extensions=self.extensions,
        )

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the DirectiveArgument's visibility in the schema.

        >>> class MyDirective(Directive, locations=[DirectiveLocation.FIELD_DEFINITION]):
        ...     name = DirectiveArgument(GraphQLNonNull(GraphQLString))
        ...
        ...     @name.visible
        ...     def name_visible(self: DirectiveArgument, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<directive_argument_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this input."""
        check_directives([directive], location=DirectiveLocation.ARGUMENT_DEFINITION)
        self.directives.append(directive)
        return self



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Literal

import pytest
from asgiref.typing import ASGIVersions
from django.contrib.auth import get_user_model
from django.test.client import MULTIPART_CONTENT, AsyncClient, Client
from graphql import FormattedExecutionResult

from undine.dataclasses import CompletedEventDC, NextEventDC
from undine.http.files import extract_files
from undine.persisted_documents.utils import to_document_id
from undine.settings import undine_settings
from undine.typing import WebSocketASGIScope

from .query_logging import capture_database_queries

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, AsyncIterator
    from http.cookies import SimpleCookie

    from django.contrib.auth.models import User
    from django.core.files import File
    from django.http import StreamingHttpResponse
    from graphql import GraphQLFormattedError

    from undine.typing import DjangoTestClientResponseProtocol

    from .query_logging import DBQueryData
    from .websocket import TestWebSocket

with suppress(ImportError):
    from .websocket import TestWebSocket


__all__ = [
    "AsyncGraphQLClient",
    "GraphQLClient",
    "GraphQLClientHTTPResponse",
    "GraphQLClientWebSocketResponse",
]


IS_CHANNELS_INSTALLED: bool = "TestWebSocket" in globals()


class WebSocketMixin:
    """Mixin to support GraphQL over WebSocket requests."""

    cookies: SimpleCookie

    async def over_websocket(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> AsyncGenerator[GraphQLClientWebSocketResponse, None]:
        """
        Send a GraphQL over WebSocket request and yield the execution results.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        websocket = self.websocket()

        if headers is not None:
            websocket.scope["headers"] += [
                (name.encode("ascii"), value.encode("ascii")) for name, value in headers.items()
            ]

        async with websocket:
            await websocket.connection_init()

            with capture_database_queries(enabled=count_queries) as queries:
                initial_result = await websocket.subscribe(body)

            if initial_result["type"] == "error":
                result = FormattedExecutionResult(data=None, errors=initial_result["payload"])
                yield GraphQLClientWebSocketResponse(result=result, database_queries=queries)
                return

            yield GraphQLClientWebSocketResponse(result=initial_result["payload"], database_queries=queries)

            while True:
                with capture_database_queries(enabled=count_queries) as queries:
                    message = await websocket.receive()

                match message["type"]:
                    case "complete":
                        return

                    case "next":
                        yield GraphQLClientWebSocketResponse(result=message["payload"], database_queries=queries)
                        continue

                    case "error":
                        result = FormattedExecutionResult(data=None, errors=message["payload"])
                        yield GraphQLClientWebSocketResponse(result=result, database_queries=queries)
                        return

                    case _:  # pragma: no cover
                        msg = f"Unexpected message type: {message['type']}"
                        raise RuntimeError(msg)

    def websocket(self, **kwargs: Any) -> TestWebSocket:
        """Create a testing websocket."""
        if not IS_CHANNELS_INSTALLED:  # pragma: no cover
            msg = "The `channels` library is not installed. Cannot create websocket."
            raise RuntimeError(msg)

        scope = self._get_websocket_scope()
        scope.update(kwargs)  # type: ignore[typeddict-item]

        return TestWebSocket(scope=scope)

    def _get_websocket_scope(self) -> WebSocketASGIScope:
        # From 'django.test.client.AsyncRequestFactory._base_scope'
        cookies = (f"{morsel.key}={morsel.coded_value}".encode("ascii") for morsel in self.cookies.values())
        path = f"/{undine_settings.WEBSOCKET_PATH}"
        return WebSocketASGIScope(  # type: ignore[typeddict-item]
            type="websocket",
            asgi=ASGIVersions(version="3.0", spec_version="1.0"),
            http_version="1.1",
            scheme="ws",
            server=("testserver", 80),
            client=("127.0.0.1", 0),
            root_path="",
            path=path,
            raw_path=path.encode("ascii"),
            query_string=b"",
            headers=[
                (b"cookie", b"; ".join(sorted(cookies))),
                (b"host", b"testserver"),
                (b"connection", b"Upgrade"),
                (b"upgrade", b"websocket"),
                (b"sec-websocket-version", b"13"),
                (b"sec-websocket-key", b"RKr31GF3kXZqsXjVT7s3Mg=="),
                (b"sec-websocket-protocol", b"graphql-transport-ws"),
            ],
            subprotocols=["graphql-transport-ws"],
            state={},
            extensions={"websocket.http.response": {}},
            cookies={k: str(v) for k, v in self.cookies.items()},
            path_remaining="",
            url_route={"args": (), "kwargs": {}},
            #
            # 'user' and 'session' are set in the `WebSocketContextManager` `AuthMiddlewareStack`.
        )  # type: ignore[typeddict-item]


class SSEMixin:
    """Mixin to support GraphQL over SSE requests."""

    cookies: SimpleCookie

    async def over_sse(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> AsyncGenerator[GraphQLClientSSEResponse, None]:
        """
        Send a GraphQL over SSE request in distinct connection mode and yield the execution results.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}
        headers = headers or {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        headers["Accept"] = "text/event-stream"

        response: StreamingHttpResponse = await self.post(  # type: ignore[assignment]
            path=f"/{undine_settings.GRAPHQL_PATH}",
            data=body,
            content_type="application/json",
            headers=headers,
        )
        responses: AsyncIterator[bytes] = aiter(response.streaming_content)

        while True:
            with capture_database_queries(enabled=count_queries) as results:
                try:
                    event_data = await anext(responses)
                except StopAsyncIteration as error:
                    msg = "Stream ended without complete event"
                    raise RuntimeError(msg) from error

            event: NextEventDC | CompletedEventDC | None = None
            for decode in [_decode_sse_next_event_dc, _decode_sse_complete_event_dc]:
                with suppress(Exception):
                    event = decode(event_data)
                    break

            if event is None:
                msg = "Unknown event type"
                raise RuntimeError(msg)

            if isinstance(event, CompletedEventDC):
                break

            yield GraphQLClientSSEResponse(event.data, results)


class GraphQLClient(WebSocketMixin, Client):
    """A GraphQL client for testing."""

    def __call__(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> GraphQLClientHTTPResponse:
        """
        Execute a GraphQL operation synchronously.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        files = extract_files(variables)
        data = _create_multipart_data(body, files) if files else body
        content_type = MULTIPART_CONTENT if files else "application/json"

        with capture_database_queries(enabled=count_queries) as results:
            response: DjangoTestClientResponseProtocol = self.post(  # type: ignore[assignment]
                path=f"/{undine_settings.GRAPHQL_PATH}",
                data=data,
                content_type=content_type,
                headers=headers,
            )

        return GraphQLClientHTTPResponse(response, results)

    def login_with_superuser(self, username: str = "admin", **kwargs: Any) -> User:
        """Create a superuser and log in as that user."""
        defaults = {
            "is_staff": True,
            "is_superuser": True,
            "email": "superuser@django.com",
            **kwargs,
        }
        user, _ = get_user_model().objects.get_or_create(username=username, defaults=defaults)
        self.force_login(user)
        return user

    def login_with_regular_user(self, username: str = "user", **kwargs: Any) -> User:
        """Create a regular user and log in as that user."""
        defaults = {
            "is_staff": False,
            "is_superuser": False,
            "email": "user@django.com",
            **kwargs,
        }
        user, _ = get_user_model().objects.get_or_create(username=username, defaults=defaults)
        self.force_login(user)
        return user


class AsyncGraphQLClient(SSEMixin, WebSocketMixin, AsyncClient):
    """An async GraphQL client for testing."""

    async def __call__(
        self,
        document: str,
        *,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
        use_persisted_document: bool = False,
        count_queries: bool = False,
    ) -> GraphQLClientHTTPResponse:
        """
        Execute a GraphQL operation asynchronously.

        :param document: GraphQL document string to send in the request.
        :param variables: GraphQL variables for the document.
        :param headers: Headers for the request.
        :param operation_name: If given document includes multiple operations,
                               this is required to select the operation to execute.
        :param use_persisted_document: Instead of sending the whole document,
                                       convert it to a persisted document ID and send that.
        :param count_queries: If True, count the number of queries executed during the request.
        """
        variables = variables or {}
        body: dict[str, Any] = {}

        if use_persisted_document:
            body["documentId"] = to_document_id(document)
        else:
            body["query"] = document

        if variables:
            body["variables"] = variables
        if operation_name is not None:
            body["operationName"] = operation_name

        files = extract_files(variables)
        data = _create_multipart_data(body, files) if files else body
        content_type = MULTIPART_CONTENT if files else "application/json"

        with capture_database_queries(enabled=count_queries) as results:
            response: DjangoTestClientResponseProtocol = await self.post(  # type: ignore[assignment]
                path=f"/{undine_settings.GRAPHQL_PATH}",
                data=data,
                content_type=content_type,
                headers=headers,
            )

        return GraphQLClientHTTPResponse(response, results)

    async def login_with_superuser(self, username: str = "admin", **kwargs: Any) -> User:
        """Create a superuser and log in as that user."""
        defaults = {
            "is_staff": True,
            "is_superuser": True,
            "email": "superuser@django.com",
            **kwargs,
        }
        user, _ = await get_user_model().objects.aget_or_create(username=username, defaults=defaults)
        await self.aforce_login(user)
        return user

    async def login_with_regular_user(self, username: str = "user", **kwargs: Any) -> User:
        """Create a regular user and log in as that user."""
        defaults = {
            "is_staff": False,
            "is_superuser": False,
            "email": "user@django.com",
            **kwargs,
        }
        user, _ = await get_user_model().objects.aget_or_create(username=username, defaults=defaults)
        await self.aforce_login(user)
        return user


class BaseGraphQLClientResponse(ABC):
    """Base class for GraphQLClient responses."""

    DB_QUERIES_NOT_ENABLED_MSG = (
        "Database queries have not been captured. "
        "Enable them using the `count_queries` parameter when calling the client."
    )

    def __init__(self, database_queries: DBQueryData) -> None:
        self._database_queries = database_queries

    @property
    @abstractmethod
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""

    # Concrete methods

    def __str__(self) -> str:
        return json.dumps(self.json, indent=2, sort_keys=True, default=str)

    def __repr__(self) -> str:
        return repr(self.json)

    def __getitem__(self, item: str) -> Any:
        return self.json[item]  # type: ignore[literal-required]

    def __contains__(self, item: str) -> bool:
        return item in self.json

    @property
    def queries(self) -> list[str]:
        """Return a list of the database queries that were executed."""
        if not self._database_queries.enabled:
            return pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        return [info.sql for info in self._database_queries.queries]

    @property
    def query_count(self) -> int:
        """Return the number of database queries that were executed."""
        if not self._database_queries.enabled:
            return pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        return self._database_queries.count

    @property
    def query_log(self) -> str:
        """Return a string representation of the database queries that were executed."""
        if not self._database_queries.enabled:
            return pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        return self._database_queries.log

    @property
    def data(self) -> dict[str, Any] | None:
        """Return the data from the response content."""
        return self.json["data"]

    @property
    def results(self) -> dict[str, Any] | list[Any] | None:
        """
        Return the results from the first top-level field in the response content.

        >>> self.json
        {"data": {"foo": {"name": "bar"}}}
        >>> self.results
        {"name": "bar"}
        """
        data = self.data or {}
        try:
            return next(iter(data.values()))
        except StopIteration:
            msg = f"No query object not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    @property
    def edges(self) -> list[dict[str, Any]]:
        """
        Return edges from the first top-level field in the response content.

        >>> self.json
        {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.edges
        [{"node": {"name": "bar"}}]
        """
        results = self.results
        if not isinstance(results, dict):
            msg = f"Edges not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

        try:
            return results["edges"]
        except (KeyError, TypeError):
            msg = f"Edges not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    def node(self, index: int) -> dict[str, Any]:
        """
        Return the node at the given index in the response content edges.

        >>> self.json
        {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.node(0)
        {"name": "bar"}
        """
        try:
            return self.edges[index]["node"]
        except (IndexError, TypeError):
            msg = f"Node {index!r} not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    @property
    def has_errors(self) -> bool:
        """Are there any errors in the response?"""
        return "errors" in self.json and self.json.get("errors") is not None

    @property
    def errors(self) -> list[GraphQLFormattedError]:
        """
        Return all errors.

        >>> self.json
        {"errors": [{"message": "bar", "path": [...], ...}]}
        >>> self.errors
        [{"message": "bar", "path": [...], ...}]
        """
        try:
            return self.json["errors"]
        except (KeyError, TypeError):
            msg = f"Errors not found in response content\nContent: {self.json}"
            return pytest.fail(msg)

    def error_message(self, selector: int | str) -> str:
        """
        Return the error message from the errors list...

        1) in the given index

        >>> self.json
        {"errors": [{"message": "bar", ...}]}
        >>> self.error_message(0)
        "bar"

        2) in the given path:

        >>> self.json
        {"errors": [{"message": "bar", "path": ["fizz", "buzz", "foo"], ...}]}
        >>> self.error_message("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["message"]
            except (IndexError, KeyError, TypeError):
                msg = f"Errors message not found from index {selector}\nContent: {self.json}"
                return pytest.fail(msg)
        else:
            try:
                return next(error["message"] for error in self.errors if error["path"][-1] == selector)
            except (StopIteration, KeyError, TypeError):
                msg = f"Errors message not found from path {selector!r}\nContent: {self.json}"
                return pytest.fail(msg)

    def assert_query_count(self, count: int) -> None:
        if not self._database_queries.enabled:
            pytest.fail(self.DB_QUERIES_NOT_ENABLED_MSG)

        if self.query_count != count:
            msg = f"Expected {count} queries, got {self.query_count}.\n{self.query_log}"
            pytest.fail(msg)


class GraphQLClientHTTPResponse(BaseGraphQLClientResponse):
    """An HTTP response from the GraphQLClient."""

    def __init__(self, response: DjangoTestClientResponseProtocol, database_queries: DBQueryData) -> None:
        self.response = response
        super().__init__(database_queries=database_queries)

    @property
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""
        return self.response.json()  # type: ignore[return-value]

    @property
    def status_code(self) -> int:
        """Return the status code of the response."""
        return self.response.status_code


class GraphQLClientWebSocketResponse(BaseGraphQLClientResponse):
    """A WebSocket response from the GraphQLClient."""

    def __init__(self, result: FormattedExecutionResult, database_queries: DBQueryData) -> None:
        self.result = result
        super().__init__(database_queries=database_queries)

    @property
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""
        return self.result


class GraphQLClientSSEResponse(BaseGraphQLClientResponse):
    """A Server-Sent Events response from the GraphQLClient."""

    def __init__(self, result: FormattedExecutionResult, database_queries: DBQueryData) -> None:
        self.result = result
        super().__init__(database_queries=database_queries)

    @property
    def json(self) -> FormattedExecutionResult:
        """Return the JSON content of the response."""
        return self.result


def _create_multipart_data(body: dict[str, Any], files: dict[File, list[str]]) -> dict[str, Any]:
    path_map: dict[str, list[str]] = {}
    files_map: dict[str, File] = {}
    for i, (file, path) in enumerate(files.items()):
        path_map[str(i)] = path
        files_map[str(i)] = file

    return {
        "operations": json.dumps(body),
        "map": json.dumps(path_map),
        **files_map,
    }


def _decode_sse_next_event_dc(event_data: bytes | str) -> NextEventDC:
    event: Literal["next"] | None = None
    data: FormattedExecutionResult | None = None

    if isinstance(event_data, bytes):
        event_data = event_data.decode()

    for part in event_data.split("\n"):
        part = part.strip()  # noqa: PLW2901
        if not part:
            continue

        key, value = part.split(":", 1)
        key = key.strip()
        value = value.strip()
        match key:
            case "event":
                if value != "next":
                    msg = f"Expected 'next' event, got {value!r}"
                    raise ValueError(msg)

                event = value

            case "data":
                # TODO: Validate data
                data = FormattedExecutionResult(**json.loads(value))

            case _:
                msg = f"Unknown key: {key!r}"
                raise ValueError(msg)

    if event is None:
        msg = "Missing 'event' key"
        raise ValueError(msg)

    if data is None:
        msg = "Missing 'data' key"
        raise ValueError(msg)

    return NextEventDC(data=data)


def _decode_sse_complete_event_dc(event_data: bytes | str) -> CompletedEventDC:
    event: Literal["complete"] | None = None

    if isinstance(event_data, bytes):
        event_data = event_data.decode()

    for part in event_data.split("\n"):
        part = part.strip()  # noqa: PLW2901
        if not part:
            continue

        key, value = part.split(":", 1)
        key = key.strip()
        value = value.strip()
        match key:
            case "event":
                if value != "complete":
                    msg = f"Expected 'complete' event, got {value!r}"
                    raise ValueError(msg)

                event = value

            case "data":
                if value:
                    msg = "Unexpected value for 'data' key"
                    raise ValueError(msg)

            case _:
                msg = f"Unknown key: {key!r}"
                raise ValueError(msg)

    if event is None:
        msg = "Missing 'event' key"
        raise ValueError(msg)

    return CompletedEventDC()



from __future__ import annotations

import asyncio
import inspect
import itertools
import sys
import types
from collections.abc import AsyncGenerator, AsyncIterable, AsyncIterator, Generator
from functools import partial
from traceback import format_tb
from types import FunctionType, GenericAlias, LambdaType, TracebackType, UnionType
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    NamedTuple,
    ParamSpec,
    Protocol,
    TypeGuard,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from asgiref.sync import sync_to_async
from graphql import GraphQLResolveInfo

from undine.dataclasses import RootAndInfoParams
from undine.exceptions import FunctionSignatureParsingError, UnionTypeMultipleTypesError
from undine.settings import undine_settings
from undine.typing import GQLInfo, LiteralArg, ParametrizedType

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Hashable, Sequence
    from types import FrameType, TracebackType

    from graphql.pyutils import AwaitableOrValue

    from undine.typing import Lambda


__all__ = [
    "FunctionEqualityWrapper",
    "as_coroutine_func_if_not",
    "async_enumerate",
    "cache_signature_if_function",
    "can_be_literal_arg",
    "cancel_awaitable",
    "get_flattened_generic_params",
    "get_instance_name",
    "get_members",
    "get_non_null_type",
    "get_root_and_info_params",
    "get_signature",
    "get_traceback",
    "get_wrapped_func",
    "has_callable_attribute",
    "is_annotated",
    "is_generic_list",
    "is_lambda",
    "is_list_of",
    "is_namedtuple",
    "is_not_required_type",
    "is_protocol",
    "is_required_type",
    "is_same_func",
    "is_subclass",
    "reverse_enumerate",
]


try:  # pragma: no cover
    from typing import is_protocol
except ImportError:  # pragma: no cover

    def is_protocol(tp: type, /) -> bool:
        """Check if the given type is a Protocol."""
        return isinstance(tp, type) and getattr(tp, "_is_protocol", False) and tp != Protocol


T = TypeVar("T")
P = ParamSpec("P")
TType = TypeVar("TType", bound=type)


def get_members(obj: object, type_: type[T]) -> dict[str, T]:
    """Get members of the given object that are instances of the given type."""
    return dict(inspect.getmembers(obj, lambda x: isinstance(x, type_)))


def get_wrapped_func(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Get the inner function of a partial function, classmethod, staticmethod, property,
    or a function wrapped with `functools.wraps`.
    """
    while True:
        if hasattr(func, "__wrapped__"):  # Wrapped with functools.wraps
            func = func.__wrapped__
            continue
        if isinstance(func, partial):
            func = func.func
            continue
        if inspect.ismethod(func) and hasattr(func, "__func__"):
            func = func.__func__
            continue
        if isinstance(func, property):
            func = func.fget
            continue
        break
    return func


def get_origin_or_noop(type_hint: T) -> T:
    """
    Get the unsubscripted version of the given type hint,
    or return the type hint itself if it's not unsubscripted.
    """
    return get_origin(type_hint) or type_hint


def get_flattened_generic_params(tp: Any) -> tuple[Any, ...]:
    """
    Get all generic parameters of the given type.
    Flattens any union types.
    """
    return tuple(a for arg in get_args(tp) for a in (get_args(arg) if isinstance(arg, UnionType) else (arg,)))


def get_non_null_type(type_: type) -> Any:
    """
    Get a non-null version of a given python type.
    If type is a Union, the Union must have exactly two members, one of which is a None.
    """
    # Remove any 'Required' or 'NotRequired' wrappers.
    bare_type = type_
    if is_required_type(type_) or is_not_required_type(type_):
        bare_type = type_.__args__[0]

    origin = get_origin(bare_type)
    if origin not in {UnionType, Union}:
        return bare_type

    args = get_flattened_generic_params(bare_type)
    if types.NoneType in args:
        args = tuple(arg for arg in args if arg is not types.NoneType)

    if len(args) > 1:
        raise UnionTypeMultipleTypesError(args=args)

    return args[0]


def cache_signature_if_function(value: Callable[..., Any], *, depth: int = 0) -> Callable[..., Any]:
    """
    Cache signature of the given value if it's a known function type.
    This allows calling `get_signature` later without knowing the function globals or locals.

    :param value: The value to cache the signature of.
    :param depth: How many function calls deep is the code calling this method compared to the parsed function?
    :returns: The "unwrapped" function, if it was a know function type, otherwise the value as is.
    """
    value = get_wrapped_func(value)
    if isinstance(value, FunctionType):
        get_signature(value, depth=depth + 1)
    return value


class _SignatureParser:
    """
    Parse the signature of a function.

    Parsed signatures are cached so that subsequent queries for the same function
    don't need to know the globals or locals of the function scope to resolve it.
    """

    def __init__(self) -> None:
        self.cache: dict[FunctionType | Callable[..., Any], inspect.Signature] = {}

    def __call__(self, func: FunctionType | Callable[..., Any], /, *, depth: int = 0) -> inspect.Signature:
        """
        Parse the signature

        :param func: The function to parse.
        :param depth: How many function calls deep is the code calling this method compared to the parsed function?
        """
        if func in self.cache:
            return self.cache[func]

        depth += 1
        frame: FrameType = sys._getframe(depth)

        # Add some common stuff to globals so that we don't encounter as many NameErrors
        # when parsing signatures if these type hints are in a `TYPE_CHECKING` block.
        extra_globals: dict[str, Any] = {
            "AsyncIterable": AsyncIterable,
            "AsyncIterator": AsyncIterator,
            "AsyncGenerator": AsyncGenerator,
            "GraphQLResolveInfo": GraphQLResolveInfo,
            "GQLInfo": GQLInfo,
            "Any": Any,
        }

        frame_locals: dict[str, Any] = frame.f_locals
        frame_globals: dict[str, Any] = frame.f_globals | extra_globals

        # Check if previous frames are in the same file and collect their locals.
        depth += 1
        prev_frame: FrameType = sys._getframe(depth)
        while prev_frame.f_code.co_filename == frame.f_code.co_filename:
            frame_locals = prev_frame.f_locals | frame_locals  # order matters!
            depth += 1
            prev_frame = sys._getframe(depth)

        try:
            sig = inspect.signature(func, eval_str=True, globals=frame_globals, locals=frame_locals)
        except NameError as error:
            raise FunctionSignatureParsingError(name=error.name, func=func) from error

        self.cache[func] = sig
        return sig


get_signature = _SignatureParser()


def has_callable_attribute(obj: object, name: str) -> bool:
    """Check if the given object has a callable attribute with the given name."""
    return hasattr(obj, name) and callable(getattr(obj, name))


def is_subclass(obj: object, cls: TType) -> TypeGuard[TType]:
    """Check if the given object is a subclass of the given class."""
    return isinstance(obj, type) and issubclass(obj, cls)  # type: ignore[arg-type]


def is_list_of(value: Any, cls: type[T], *, allow_empty: bool = False) -> TypeGuard[list[T]]:
    """
    Check if the value is a homogeneous list of the given type.
    List must have at least one item to be considered a list of the given type, unless `allow_empty` is True.
    """
    max_length = 0 if allow_empty else 1
    return isinstance(value, list) and len(value) >= max_length and all(isinstance(item, cls) for item in value)  # type: ignore[arg-type]


def is_generic_list(type_: type) -> TypeGuard[GenericAlias]:
    """Check if the given type is a generic list, i.e., `list[str]`."""
    return isinstance(type_, GenericAlias) and issubclass(type_.__origin__, list)  # type: ignore[arg-type]


def is_lambda(func: Callable[..., Any]) -> TypeGuard[Lambda]:
    """Check if the given function is a lambda function."""
    return isinstance(func, LambdaType) and func.__name__ == "<lambda>"


def is_required_type(type_: Any) -> TypeGuard[ParametrizedType]:
    """Check if the given type is a TypedDict `Required` type."""
    return isinstance(type_, ParametrizedType) and getattr(type_.__origin__, "_name", None) == "Required"  # type: ignore[misc]


def is_not_required_type(type_: Any) -> TypeGuard[ParametrizedType]:
    """Check if the given type is a TypedDict `Required` type."""
    return isinstance(type_, ParametrizedType) and getattr(type_.__origin__, "_name", None) == "NotRequired"  # type: ignore[misc]


def is_namedtuple(obj: Any) -> TypeGuard[type[NamedTuple]]:
    """Check if the given object is a namedtuple class or not."""
    return (
        is_subclass(obj, tuple)
        and getattr(obj, "_fields", None) is not None
        and getattr(obj, "_field_defaults", None) is not None
        and getattr(obj, "_asdict", None) is not None
    )


def is_annotated(obj: Any) -> TypeGuard[Annotated]:
    """Check if the given object is an `Annotated` instance or not."""
    return get_origin_or_noop(obj) is Annotated


def is_same_func(func_1: FunctionType | Callable[..., Any], func_2: FunctionType | Callable[..., Any], /) -> bool:
    """
    Check if the given functions are the same function.
    Handles partial functions and functions wrapped with `functools.wraps`.
    """
    return get_wrapped_func(func_1) == get_wrapped_func(func_2)


def can_be_literal_arg(key: Any) -> TypeGuard[LiteralArg]:
    return isinstance(key, LiteralArg)


def get_instance_name() -> str:
    """
    Perform some python black magic to find the name of the variable
    to which an instance of a class is being assigned to.
    Should be used in the '__init__' method.

    Note: This only works if the instance initializer is called on the
    same line as the variable for it's defined to.
    """
    frame = sys._getframe(2)
    source = inspect.findsource(frame)[0]
    line = source[frame.f_lineno - 1]
    definition = line.split("=", maxsplit=1)[0]
    return definition.split(":", maxsplit=1)[0].strip()


class FunctionEqualityWrapper(Generic[T]):
    """
    Adds equality checks for a function based on the provided context.
    Function is equal to another function if it's also wrapped with this class
    and the provided contexts are equal.
    """

    def __init__(self, func: Callable[[], T], context: Hashable) -> None:
        self.func = func
        self.context = context

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.context == other.context

    def __hash__(self) -> int:
        return hash(self.context)

    def __call__(self) -> T:
        return self.func()


def get_root_and_info_params(func: FunctionType | Callable[..., Any], *, depth: int = 0) -> RootAndInfoParams:
    """
    Inspect the function signature to figure out which parameters are
    the root and info parameters of a GraphQL resolver function, if any.

    `root_param` is the first parameter of the function if it's named `self`, `cls`,
    or the name configured with the `RESOLVER_ROOT_PARAM_NAME` setting (`root` by default).

    `info_param` is annotated as `GraphQLResolveInfo` or `GQLInfo` if it exists,
    and is usually the second parameter of the function, but can also be in any
    other position other than first.
    """
    sig = get_signature(func, depth=depth + 1)

    root_param: str | None = None
    info_param: str | None = None
    for i, param in enumerate(sig.parameters.values()):
        if i == 0 and param.name in {"self", "cls", undine_settings.RESOLVER_ROOT_PARAM_NAME}:
            root_param = param.name

        elif get_origin_or_noop(param.annotation) in {GraphQLResolveInfo, GQLInfo}:
            info_param = param.name
            break

    return RootAndInfoParams(root_param=root_param, info_param=info_param)


def reverse_enumerate(sequence: Sequence[T]) -> Generator[tuple[int, T], None, None]:
    """
    Enumerate the given sequence in reverse order.
    Using this allows using `.pop(index)` on the sequence to remove the iterated item if needed.

    >>> x = ["a", "b", "c"]
    >>> for i, item in reverse_enumerate(x):
    >>>     x.pop(i)
    """
    for index in range(len(sequence) - 1, -1, -1):
        yield index, sequence[index]


async def async_enumerate(it: AsyncIterable[T]) -> AsyncGenerator[tuple[int, T], None]:
    """Enumerate the given async iterable."""
    counter = itertools.count()
    async for item in it:
        yield next(counter), item


def get_traceback(traceback: TracebackType) -> list[str]:
    """Format the given traceback into a list of strings."""
    return [subline for line in format_tb(traceback) for subline in line.split("\n")]


def as_coroutine_func_if_not(func: Callable[P, AwaitableOrValue[T]], /) -> Callable[P, Awaitable[T]]:
    """Convert function to a coroutine function using sync_to_async if needed."""
    if inspect.iscoroutinefunction(func):
        return func
    return sync_to_async(func)  # type: ignore[arg-type]


def cancel_awaitable(value: Awaitable) -> None:
    """Cancel the given awaitable."""
    if isinstance(value, types.CoroutineType):
        value.close()
        return

    if isinstance(value, asyncio.Future):
        value.cancel()
        return


class delegate_to_subgenerator:  # noqa: N801
    """
    Allows delegating how a generator exists to a subgenerator.

    >>> def subgenerator():
    ...     for _ in range(2):
    ...         yield
    >>>
    >>> def generator():
    >>>     with delegate_to_subgenerator(subgenerator()) as sub:
    ...         for _ in sub:
    ...             yield
    >>>
    >>> for item in generator():
    ...     pass

    If the generator exists normally, the subgenerator will be closed.
    If the generator exists with an exception, the error is propagated to the subgenerator
    so that it may handle the error.
    """

    def __init__(self, gen: Generator[None, None, None] | AsyncGenerator[None, None]) -> None:
        """
        Allows delegating how a generator exists to a subgenerator.

        :param gen: The generator to delegate to. If generator is an async generator,
                    must use `async with` syntax to delegate. For regular generators,
                    plain `with` syntax must be used.
        """
        self.gen = gen

    def __enter__(self) -> Generator[None, None, None]:
        if not isinstance(self.gen, Generator):
            msg = "Given object is not a Generator"
            raise TypeError(msg)

        return self.gen

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        if not isinstance(self.gen, Generator):  # type: ignore[unreachable]
            msg = "Given object is not a Generator"
            raise TypeError(msg)

        # If no exception was raised, close the generator.
        if exc_type is None:
            self.gen.close()
            return False

        # Otherwise, allow the subgenerator to handle the exception.
        # This has mostly been copied from `contextlib._GeneratorContextManager.__exit__`.
        if exc_value is None:
            exc_value = exc_type()

        try:
            self.gen.throw(exc_value)

        except StopIteration as error:
            return error is not exc_value

        except RuntimeError as error:
            if error is exc_value:
                error.__traceback__ = traceback
                return False
            if isinstance(exc_value, StopIteration) and error.__cause__ is exc_value:
                exc_value.__traceback__ = traceback
                return False
            raise

        except BaseException as error:
            if error is not exc_value:
                raise
            error.__traceback__ = traceback
            return False

        try:
            msg = "generator didn't stop after throw()"
            raise RuntimeError(msg)
        finally:
            self.gen.close()

    async def __aenter__(self) -> AsyncGenerator[None, None]:
        if not isinstance(self.gen, AsyncGenerator):
            msg = "Given object is not an AsyncGenerator"
            raise TypeError(msg)

        return self.gen

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        if not isinstance(self.gen, AsyncGenerator):
            msg = "Given object is not an AsyncGenerator"
            raise TypeError(msg)

        # If no exception was raised, close the generator.
        if exc_type is None:
            await self.gen.aclose()
            return False

        # Otherwise, allow the subgenerator to handle the exception.
        # This has mostly been copied from `contextlib._AsyncGeneratorContextManager.__aexit__`.
        if exc_value is None:
            exc_value = exc_type()

        try:
            await self.gen.athrow(exc_value)

        except StopAsyncIteration as error:
            return error is not exc_value

        except RuntimeError as error:
            if error is exc_value:
                error.__traceback__ = traceback
                return False
            if isinstance(exc_value, (StopIteration, StopAsyncIteration)) and error.__cause__ is exc_value:
                exc_value.__traceback__ = traceback
                return False
            raise

        except BaseException as error:
            if error is not exc_value:
                raise
            error.__traceback__ = traceback
            return False

        try:
            msg = "generator didn't stop after athrow()"
            raise RuntimeError(msg)
        finally:
            await self.gen.aclose()



from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import F, Model, Q, QuerySet
from graphql import GraphQLList, GraphQLNonNull, GraphQLType

from undine import Calculation, InterfaceField, MutationType, QueryType, UnionType
from undine.converters import is_many
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.exceptions import ModelFieldDoesNotExistError, ModelFieldNotARelationOfModelError
from undine.pagination import OffsetPagination
from undine.parsers import parse_return_annotation
from undine.relay import Connection
from undine.typing import CombinableExpression, ModelField
from undine.utils.model_utils import get_model_field
from undine.utils.reflection import get_non_null_type, get_origin_or_noop


@is_many.register
def _(ref: ModelField, **kwargs: Any) -> bool:
    return bool(ref.many_to_many) or bool(ref.one_to_many)


@is_many.register
def _(ref: type[Model], **kwargs: Any) -> bool:
    try:
        field = get_model_field(model=kwargs["model"], lookup=kwargs["name"])
    except ModelFieldDoesNotExistError:
        return False

    if field.related_model != ref:
        raise ModelFieldNotARelationOfModelError(field=field.name, model=kwargs["model"], related=ref)

    return is_many(field, **kwargs)


@is_many.register
def _(ref: TypeRef, **kwargs: Any) -> bool:
    ann = get_non_null_type(ref.value)
    annotation = get_origin_or_noop(ann)
    return isinstance(annotation, type) and issubclass(annotation, list | set | tuple | QuerySet)


@is_many.register
def _(ref: CombinableExpression, **kwargs: Any) -> bool:
    return is_many(ref.output_field)


@is_many.register
def _(_: F | Q, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(ref: LazyRelation, **kwargs: Any) -> bool:
    return is_many(ref.field)


@is_many.register
def _(_: LazyGenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(ref: type[Calculation], **kwargs: Any) -> bool:
    return is_many(TypeRef(value=ref.__returns__))


@is_many.register
def _(ref: GraphQLType, **kwargs: Any) -> bool:
    return isinstance(ref, GraphQLList) or (isinstance(ref, GraphQLNonNull) and isinstance(ref.of_type, GraphQLList))


@is_many.register
def _(ref: FunctionType, **kwargs: Any) -> bool:
    ann = get_non_null_type(parse_return_annotation(ref))
    annotation = get_origin_or_noop(ann)
    return isinstance(annotation, type) and issubclass(annotation, list | set | tuple | QuerySet)


@is_many.register
def _(_: type[QueryType], **kwargs: Any) -> bool:
    field = get_model_field(model=kwargs["model"], lookup=kwargs["name"])
    return is_many(field, **kwargs)


@is_many.register
def _(_: type[MutationType], **kwargs: Any) -> bool:
    field = get_model_field(model=kwargs["model"], lookup=kwargs["name"])
    return is_many(field, **kwargs)


@is_many.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_many.register
def _(_: Connection, **kwargs: Any) -> bool:
    # Connection edges have many nodes, but the connection itself is a single node.
    return False


@is_many.register
def _(_: OffsetPagination, **kwargs: Any) -> bool:
    return True


@is_many.register
def _(_: type[UnionType], **kwargs: Any) -> bool:
    # You always want multiple results from a union,
    # otherwise you should know its type beforehand.
    return True


@is_many.register
def _(_: InterfaceField, **kwargs: Any) -> bool:
    field = get_model_field(model=kwargs["model"], lookup=kwargs["name"])
    return is_many(field, **kwargs)



from __future__ import annotations

from django.core.files import File
from django.core.files.uploadedfile import UploadedFile
from django.db.models.fields.files import ImageFieldFile

from undine.utils.validators import validate_image_upload, validate_image_url

from ._definition import ScalarType

__all__ = [
    "GraphQLImage",
    "image_scalar",
]


image_scalar: ScalarType[UploadedFile, str] = ScalarType(
    name="Image",
    description="Represents an image file.",
)

GraphQLImage = image_scalar.as_graphql_scalar()


@image_scalar.parse.register
def _(value: UploadedFile) -> UploadedFile:
    validate_image_upload(value)
    return value


@image_scalar.serialize.register
def _(value: ImageFieldFile) -> str:
    return value.url


@image_scalar.serialize.register
def _(value: File) -> str:
    return value.name or ""


@image_scalar.serialize.register
def _(value: str) -> str:
    validate_image_url(value)
    return value



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib import admin

from .models import PersistedDocument

if TYPE_CHECKING:
    from django.http import HttpRequest


__all__ = [
    "PersistedDocumentAdmin",
]


@admin.register(PersistedDocument)
class PersistedDocumentAdmin(admin.ModelAdmin):
    list_display = ("document_id", "created_at")
    search_fields = ("document_id", "document")
    fields = ("document_id", "document", "created_at")
    readonly_fields = ("document_id", "document", "created_at")

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False

    def has_change_permission(self, request: HttpRequest, obj: PersistedDocument | None = None) -> bool:
        return False

    def has_delete_permission(self, request: HttpRequest, obj: PersistedDocument | None = None) -> bool:
        return False



from __future__ import annotations

from types import FunctionType, NoneType, UnionType
from typing import Any, get_origin

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.db.models import F, Field, ManyToManyRel, ManyToOneRel, OneToOneRel, Q
from graphql import GraphQLNonNull, GraphQLType

from undine import Calculation, InterfaceField, QueryType
from undine import Field as UndineField
from undine.converters import is_field_nullable
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.pagination import OffsetPagination
from undine.parsers import parse_return_annotation
from undine.relay import Connection
from undine.typing import CombinableExpression
from undine.utils.model_utils import get_model_field
from undine.utils.reflection import get_flattened_generic_params


@is_field_nullable.register
def _(ref: Field, **kwargs: Any) -> bool:
    return getattr(ref, "null", False)


@is_field_nullable.register
def _(_: OneToOneRel, **kwargs: Any) -> bool:
    return True


@is_field_nullable.register
def _(_: ManyToOneRel | ManyToManyRel, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(ref: CombinableExpression, **kwargs: Any) -> bool:
    return is_field_nullable(ref.output_field, **kwargs)


@is_field_nullable.register
def _(_: F, **kwargs: Any) -> bool:
    caller: UndineField = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return is_field_nullable(field, **kwargs)


@is_field_nullable.register
def _(_: Q, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(ref: LazyRelation, **kwargs: Any) -> bool:
    return is_field_nullable(ref.field, **kwargs)


@is_field_nullable.register
def _(_: LazyGenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(ref: TypeRef, **kwargs: Any) -> bool:
    origin = get_origin(ref.value)
    if origin is not UnionType:
        return False
    return NoneType in get_flattened_generic_params(ref.value)


@is_field_nullable.register
def _(ref: type[Calculation], **kwargs: Any) -> bool:
    return is_field_nullable(TypeRef(value=ref.__returns__))


@is_field_nullable.register
def _(ref: GraphQLType, **kwargs: Any) -> bool:
    return not isinstance(ref, GraphQLNonNull)


@is_field_nullable.register
def _(ref: FunctionType, **kwargs: Any) -> bool:
    annotation = parse_return_annotation(ref)
    if not isinstance(annotation, UnionType):
        return False
    return NoneType in get_flattened_generic_params(annotation)


@is_field_nullable.register
def _(_: type[QueryType], **kwargs: Any) -> bool:
    caller: UndineField = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return is_field_nullable(field, **kwargs)


@is_field_nullable.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(_: GenericRelation, **kwargs: Any) -> bool:
    # Reverse relations are always nullable (Django can't enforce that a
    # foreign key on the related model points to this model).
    return True


@is_field_nullable.register
def _(_: Connection, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(_: OffsetPagination, **kwargs: Any) -> bool:
    return False


@is_field_nullable.register
def _(_: InterfaceField, **kwargs: Any) -> bool:
    caller: UndineField = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return is_field_nullable(field, **kwargs)
